#!/usr/bin/python

import os
import sys
import logging
import time
import optparse
import urllib
import StringIO
import struct
import tempfile
import traceback

from SimpleXMLRPCServer import SimpleXMLRPCServer

sys.path.append(os.path.abspath('cs'))
sys.path.append(os.path.abspath(os.path.join('..', 'cs')))
sys.path.append(os.path.abspath('.'))

# print '\n'.join(sys.path)

from ncrypt0.rsa import RSAKey, RSAError
from ncrypt0.digest import DigestType, Digest
from nitro.selectreactor import SelectReactor
from nitro.http import HttpRequest
from nitro.tcp import tcpConnect, TCPStream, TCPMessageStream
from nitro.async import AsyncOp
from cspace.main.service import CSpaceService
from cspace.main.common import isValidUserName, localSettings, profileSettings, LocalSettings
from cspace.main.profile import Profile, Contact, createProfile, listProfiles, loadProfile, saveProfileContacts, clearProfileContacts
from cspace.main.appletserver import AppletServer
from cspace.network.sslutil import set_CA_cert_path, set_CA_key_path

#------------------------------------------------------------------------------ 

reactor = None

# CSpace can open a socket for us - directly to the remote machine
# after we sent a file we keep it opened ...
# to be able to send the next file as soon as possible
# we can send several files at same time - several opened sockets to this user
# of course we want to close the socket after sent ...
# but let's keep last socket opened ...    or last N sockets 
OPENED_SENDING_STREAMS_NUMBER = 1
MAXIMUM_SENDING_STREAMS_NUMBER = 3
# after we receive a file from this man
# we check if we have more files receiving at the moment from him
# if yes - close this stream because not needed
# so keep last N streams opened after receiving
# may be we want to have more streams opened to increase speed 
OPENED_RECEIVING_STREAMS_NUMBER = 1

_AppLogger = None
_TempLocation = None
_Notifier = None
_ReceivingStreams = {}
_SendingStreams = {}
_SendingDelay = 0.01
_OutgoingFilesDict = {}
_InActionFilesDict = {}
_ContactStatus = {}


#------------------------------------------------------------------------------ 

class Application:
    def __init__(self, reactor, options, args):
        global _TempLocation
        
        self.reactor = reactor

        if options.homedir:
            self.homeDir = options.homedir.strip('"')
        else:
            self.homeDir = os.path.join("~", ".CSpace")
            self.homeDir = os.path.expanduser(self.homeDir)
        if not os.path.isdir(self.homeDir):
            try:
                os.makedirs(self.homeDir)
            except:
                logger().exception('can not create %s folder' % self.homeDir)
        os.environ['CSPACEHOMEDIR'] = self.homeDir

        if options.tmpdir:
            _TempLocation = options.tmpdir.strip('"')
        
        self.request_registration = False
        if options.register:
            self.request_registration = True
            if len(args) >= 2:
                self.username = args[0]
                self.password = args[1]
            else:
                import time
                self.username = 'dhn' + time.strftime('%Y%m%d%H%M%S')
                self.password = os.urandom(12)
        
        self.probeDict = {}
        self.xmlport = options.xmlport
        self.seedNodes = []
        for node in options.seednodes:
            n = node.split(':')
            self.seedNodes.append((n[0], int(n[1])))
        
    def init(self):
        self.service = CSpaceService(self.seedNodes, reactor=self.reactor)
        notify()
        self.server = SimpleXMLRPCServer(('localhost', self.xmlport), allow_none=True, logRequests=False)
        self.reactor.addReadCallback(self.server, self.server.handle_request)
        self.server.register_instance(self, allow_dotted_names = True)
        self.xmlrpcport = self.server.socket.getsockname()[1]
        self.appletport = self.service.appletServer.listenPort
        
        ca_key_path = os.path.join(self.homeDir, 'ca.key')
        ca_cert_path = os.path.join(self.homeDir, 'ca.pem')
        try:
            fout = open(ca_key_path, 'wb')
            fout.write(CA_KEY().decode('ascii'))
            fout.close()
            fout = open(ca_cert_path, 'wb')
            fout.write(CA_CERT().decode('ascii'))
            fout.close()
        except:
            logger().exception('error creating CA files')
        set_CA_key_path(ca_key_path)
        set_CA_cert_path(ca_cert_path)
        
        self.dispatcher = self.service.dispatcher
        self.dispatcher.register('service.start', self.onStarted)
        self.dispatcher.register('service.stop', self.onStopped)
        self.dispatcher.register('profile.connecting', self.onConnecting)
        self.dispatcher.register('profile.disconnecting', self.onDisconnecting)
        self.dispatcher.register('profile.online', self.onOnline)
        self.dispatcher.register('profile.offline', self.onOffline)
        self.dispatcher.register('profile.reconnecting', self.onReconnect)
        self.dispatcher.register('contact.online', self.onUserOnline)
        self.dispatcher.register('contact.action', self.onUserAction)
        self.dispatcher.register('contacts.probing', self.onContactsProbing)
        self.dispatcher.register('contacts.probed', self.onContactsProbed)
        self.dispatcher.register('contact.offline', self.onUserOffline)
        self.dispatcher.register('offlineim.get', self.onOfflineReceive)
        self.dispatcher.register('offlineim.put', self.onOfflineStore)
        self.dispatcher.register('offlineim.del', self.onOfflineDelete)
        self.dispatcher.register('error', self.onError)

        actionId = self.registerSendAction()
        logger().info('registered "send" action with ID %s' % str(actionId))

        self.registerReceiveService()
        logger().info('registered "receive" service')

        self.ProcessSending()

    def ProcessSending(self):
        global _SendingDelay
        global _SendingStreams
        global _ReceivingStreams
        global _InActionFilesDict
        global MAXIMUM_SENDING_STREAMS_NUMBER
        keyID = self.readOutgoingQueue()
        if keyID:
            _SendingDelay = 0.01
        else:
            if _SendingDelay < 2.0:
                _SendingDelay *= 2.0
        # attenuation
        self.reactor.callLater(_SendingDelay, self.ProcessSending)
        self.closeSendingStreams()
        if not keyID:
            return
        if not _InActionFilesDict.has_key(keyID):
            _InActionFilesDict[keyID] = []
        if not _SendingStreams.has_key(keyID):
            _SendingStreams[keyID] = []
        #--- HERE IS SENDING A FILE TO REMOTE GUY ---
        # logger().info('found some file for %s, we have %d opened streams with him and %d actions' % (keyID, len(_SendingStreams[keyID]), len(_InActionFilesDict[keyID])))
        if _SendingStreams.has_key(keyID) and len(_SendingStreams[keyID]) > 0:
            for stream in _SendingStreams[keyID]:
                if stream.state == 'ready':
                    # logger().info('using existing stream to send the file to %s ' % keyID)
                    filename, transferID, description = popFileName(keyID)
                    if filename == '':
                        notify().sent(keyID, filename, transferID, 'failed', None, 'no files to send in the queue')
                        return
                    stream.newFile(filename, transferID, description)
                    return
        if _SendingStreams.has_key(keyID) and len(_SendingStreams[keyID]) + len(_InActionFilesDict[keyID]) >= MAXIMUM_SENDING_STREAMS_NUMBER:
            # logger().info('too many sending streams at the moment with %s ' % keyID)
            return
        if len(_InActionFilesDict[keyID]) > 0:
            # logger().info('already started action with %s ' % keyID)
            return
        filename, transferID, description = popFileName(keyID)
        if filename == '':
            notify().sent(keyID, filename, transferID, 'failed', None, 'no files to send in the queue')
            return
        _InActionFilesDict[keyID].append((filename, transferID, description))
        self.service.action(keyID, 'send')

    def readOutgoingQueue(self):
        global _OutgoingFilesDict
        for keyid in _OutgoingFilesDict.keys():
            if len(_OutgoingFilesDict[keyid]) > 0:
                return str(keyid)
        return None

    def closeSendingStreams(self):
        global _SendingStreams
        global OPENED_SENDING_STREAMS_NUMBER
        to_close = []
        for keyID, streams in _SendingStreams.items():
            for stream in streams:
                if stream.state != 'ready':
                    continue
                if time.time() - stream.ready > 60 or OPENED_SENDING_STREAMS_NUMBER == 0:
                    to_close.append(stream)
        while len(to_close) > 0:
            stream = to_close.pop(0)
            if len(_SendingStreams[keyID]) > OPENED_SENDING_STREAMS_NUMBER or OPENED_SENDING_STREAMS_NUMBER == 0:
                stream.connectionLost()
        del to_close
    
    #------------------------------------------------------------------------------ 

    def run(self):
        try:
            self.writePID(self.xmlrpcport, self.appletport)
            self.service.run()
            self.deletePID()
        except:
            logger().exception('ERROR!')
    
    def register(self, username, password):
        logger().info('register %s' % username)
        rsaKey = RSAKey()
        rsaKey.generate( bits=2048 )
        data = 'username:%s' % username
        digestType = DigestType('SHA1')
        digest = Digest(digestType).digest( data )
        signature = rsaKey.sign( digest, digestType )
        form = dict(username=username, public_key=rsaKey.toPEM_PublicKey(), signature=signature)
        postData = urllib.urlencode( form )
        self._reactor = SelectReactor()
        request = HttpRequest(self._reactor)
        httpOp = request.post('http://identity.datahaven.net/cspacekey', postData, self.onRegisterResponse )
        self._keyId = -1
        self._reactor.run()
        if self._keyId >= 0:
            st = self.settings()
            st.setString('Settings/SavedProfile', username)
            st.setString('Settings/SavedPassword', password)
            st.setInt('Settings/RememberKey', 1)
            self.createProfile(rsaKey, password, username, str(self._keyId))
            logger().info(username + ' registered, key id is ' + str(self._keyId))
        else:
            logger().info('registration failed')
        del self._keyId
        del self._reactor

    def sendControlCommand(self, command, data=''):
        controlCommand(command, data)

    def getPIDpath(self, appName="CSpace"):
        pidfile = "%s.run" % appName
        pidpath = self.homeDir
        if not os.path.exists(pidpath):
            os.makedirs(pidpath)
        pidpath = os.path.join(pidpath, pidfile)
        return pidpath

    def writePID(self, xmlrpcport, appletport):
        pidpath = self.getPIDpath()
        pidfile = file(pidpath, "wb")
        pidfile.write("%i\n%i\n%i\n" % (os.getpid(), xmlrpcport, appletport))
        pidfile.close()
    
    def deletePID(self):
        pidpath = self.getPIDpath()
        os.unlink(pidpath)

    def settings(self, location='CSpace'):
        return localSettings(location)
    
    def profile(self, settings_location='CSpace', profile_location='CSpaceProfiles'):
        if not self.settings(settings_location).getInt('Settings/RememberKey', 0):
            return None
        if not self.settings(settings_location).getInt('Settings/RememberKey', 0):
            return None
        entries = [entry for userName,keyId,entry in listProfiles(profile_location)]
        entry = self.settings(settings_location).getString('Settings/SavedProfile')
        password = self.settings(settings_location).getString('Settings/SavedPassword')
        if not (entry and password and (entry in entries)):
            return None
        return loadProfile(entry, password, profile_location)        

    def createProfile(self, rsaKey, password, userName, keyId, profile_location='CSpaceProfiles') :
        ps = profileSettings(profile_location)
        baseEntry = userName
        entry = baseEntry
        suffix = 0
        while ps.getData(entry+'/PrivateKey') :
            suffix += 1
            entry = '%s-%d' % (baseEntry,suffix)
        encKey = rsaKey.toPEM_PrivateKey( password )
        ps.setData(entry+'/PrivateKey', encKey)
        ps.setData(entry+'/Name', userName)
        if keyId is not None :
            ps.setData(entry+'/KeyID', keyId)
        profile = Profile(rsaKey, userName, keyId, entry)
        return profile

    def patchPermissions(self):
        import cspace.main.permissions        
        cspace.main.permissions.predefinedPermissions = "allow <any-user> service <any-service>\n"
        
    def setPermissions(self):
        self.service.session.permissions.setUserPermissions("allow <any-user> service receive\n")

    def registerSendAction(self):
        self.service.actionManager.activeActions.append(('send', 'send', '', 0))
        actionID = self.service.actionManager.registerAction('send', self.onAction, 0)
        self.service.appletServer.actions.append(actionID)
        self.service.actionManager.setDefaultAction(actionID)
        return actionID

    def registerReceiveService(self):
        logger().info('registerReceiveService callback id is %d' % id(self.onService))
        return self.service.session.registerService('receive', self.onService)

    def isContact(self, keyID):
        if self.service.profile is None:
            return False
        return keyID in self.service.profile.contactNames.keys()

    def contact(self, keyID):
        if self.service.profile is None:
            return None
        if not self.isContact(keyID):
            return None
        return self.service.profile.getContactByName(keyID)

    def state(self, keyID):
        global _ContactStatus
        return _ContactStatus.get(keyID, 'unknown')

    def status(self, keyID):
        c = self.contact(keyID)
        if c is None:
            return self.service.UNKNOWN
        return hasattr(c, 'status') and c.status or self.service.UNKNOWN
    
    #------------------------------------------------------------------------------ 

    def online(self):
        return self.service.online()
        
    def offline(self):
        return self.service.offline()
        
    def remove(self, buddy):
        return self.service.remove(buddy)
        
    def list(self):
        return self.service.list()
        
    def probe(self, keyID, cb=None):
        if self.service.status != self.service.ONLINE:
            if cb:
                cb(self.contact(keyID), 'offline')
            return 
        if not self.probeDict.has_key(keyID):
            self.probeDict[keyID] = []
        if cb:
            self.probeDict[keyID].append(cb)
        ret = self.service.probe(keyID)
        if ret != True:
            for cb in self.probeDict[keyID]:
                cb(self.contact(keyID), 'offline')
            del self.probeDict[keyID]
        return ret
        
    def buddy(self, buddy):
        return self.service.contactinfo(buddy)

    def stop(self):
        return self.service.stop()
        
    def action(self, keyID, action):
        return self.service.action(keyID, action)

    def send(self, keyID, filename, transferID, description):
        def _do_send():
            if self.status(keyID) == self.service.ONLINE:
                pushFileName(str(keyID), filename, transferID, description)
            else:
                notify().sent(keyID, filename, transferID, 'failed', None, 'contact is offline')
        def _probed(contact, status):
            _do_send()
        def _do_probe():
            if self.probeDict.has_key(keyID) and len(self.probeDict[keyID]) > 0:
                self.probeDict[keyID].append(_probed)
            else: 
                # logger().info('probe %s status' % keyID)
                self.probe(str(keyID), _probed)
        def _added(contact):
            if contact is None:
                return
            _do_probe()
        if self.service.profile is None:
            logger().info('send ERROR, profile is None')
            notify().sent(keyID, filename, transferID, 'failed', None, 'profile is None')
            return False
        if not self.isContact(keyID):
            logger().info('add %s to contacts' % keyID)
            self.add(keyID, _added)
            return True
        if self.status(keyID) != self.service.ONLINE:
            _do_probe()
            return True
        _do_send()
        return True

    def cancel(self, transferID):
        global _OutgoingFilesDict
        global _SendingStreams
        for keyid in _OutgoingFilesDict.keys():
            for i in range(len(_OutgoingFilesDict[keyid])):
                filename, transfer_id, description = _OutgoingFilesDict[keyid][i]
                if transfer_id == transferID:
                    del _OutgoingFilesDict[keyid][i]
                    notify().sent(keyid, filename, transferID, 'failed', None, 'file transfer canceled')
                    break
        for keyid in _SendingStreams.keys():
            for i in range(len(_SendingStreams[keyid])):
                stream = _SendingStreams[keyid][i]
                if stream.transferID and stream.transferID == transferID:
                    stream.connectionLost()
                    break 
        
    def add(self, keyID, add_callback=None):
        if self.service.profile.contactNames.has_key(keyID):
            if add_callback is not None:
                add_callback(self.service.profile.contactNames.get(keyID))
            return
        
        def _onLookupResponse(responseCode, data):
            if responseCode != 200 :
                if add_callback is not None:
                    add_callback(None)
                return
            inp = StringIO.StringIO( data )
            name = inp.readline().strip()
            pemPublicKey = inp.read()
            if name and not isValidUserName(name) :
                if add_callback is not None:
                    add_callback(None)
                return

            k = RSAKey()
            try:
                k.fromPEM_PublicKey( pemPublicKey )
            except:
                logger().exception('can not create a key for user %s, keyID=%s' %(name, keyID))
                if add_callback is not None:
                    add_callback(None)

            contact = Contact(k, str(keyID))
            if str(keyID) in self.service.profile.contactNames.keys():
                self.service.profile.removeContact(contact)
            self.service.profile.addContact(contact)
            saveProfileContacts(self.service.profile)
            if add_callback is not None:
                add_callback(contact)

        httpOp = HttpRequest( self.reactor )
        httpOp.get('http://identity.datahaven.net/cspacekeys/%s' % keyID, _onLookupResponse )
        
    def set_sending_streams(self, num):
        global OPENED_SENDING_STREAMS_NUMBER    
        OPENED_SENDING_STREAMS_NUMBER = int(num)
        logger().info('set sending streams number to %d' % int(num))
        
    def set_receiving_streams(self, num):
        global OPENED_RECEIVING_STREAMS_NUMBER
        OPENED_RECEIVING_STREAMS_NUMBER = int(num)
        logger().info('set receiving streams number to %d' % int(num))
        
    #------------------------------------------------------------------------------ 

    def onRegisterResponse( self, returnCode, data ) :
        logger().info('onRegisterResponse retCode=%s' % str(returnCode))
        if returnCode != 200 :
            self._keyId = -1
        else:
            try :
                keyId = int(data)
                self._keyId = keyId
            except:
                self._keyId = -1
        self._reactor.stop()

    def onError(self, msg):
        self.sendControlCommand('error', '"%s"' % msg)

    def onSessionService(self):
        pass

    def onStarted(self):
        self.sendControlCommand('service-started')
        
    def onConnecting(self, profile):
        self.sendControlCommand('service-connecting')

    def onOnline(self, profile):
        self.sendControlCommand('online')
        if self.service.profile:
            self.setPermissions()
            # clearProfileContacts(self.service.profile)
            # self.service.profile.contactKeys.clear()
            # self.service.profile.contactNames.clear()
            # logger().info('all contacts was removed')
    
    def onDisconnecting(self, profile):
        self.sendControlCommand('disconnecting')

    def onOffline(self, profile):
        self.sendControlCommand('offline')

    def onReconnect(self, e):
        self.sendControlCommand('service-reconnect')

    def onStopped(self):
        global _SendingStreams
        global _ReceivingStreams
        for streams in _SendingStreams.values():
            for stream in streams:
                stream.connectionLost()
        for streams in _ReceivingStreams.values():
            for stream in streams:
                stream.connectionLost()
        self.sendControlCommand('service-stopped')
    
    def onContactsProbing(self, contacts):
        # self.sendControlCommand('probing')
        pass
    
    def onContactsProbed(self, contacts):
        # self.sendControlCommand('probed')
        pass
        
    def onUserOnline(self, contact):
        global _ContactStatus
        if self.probeDict.has_key(contact.name):
            for cb in self.probeDict[contact.name]:
                cb(contact, 'online')
            del self.probeDict[contact.name]
        _ContactStatus[contact.name] = 'online'
        notify().contact_status(contact.name, 'online')

    def onUserOffline(self, contact):
        global _ContactStatus
        if self.probeDict.has_key(contact.name):
            for cb in self.probeDict[contact.name]:
                cb(contact, 'offline')
            del self.probeDict[contact.name]
        _ContactStatus[contact.name] = 'offline'
        notify().contact_status(contact.name, 'offline')
        
    def onUserAction(self, contact, action, retval):
        pass
    
    def onOfflineReceive(self, ts, msg, peerPubKey):
        pass
        
    def onOfflineStore(self, msg):
        pass
        
    def onOfflineDelete(self, envelope):
        pass
        
    def onAction(self, contactName):
        keyid = str(contactName)
        def onConnectUser( err, sock ) :
            global _SendingStreams
            global _InActionFilesDict 
            if not _InActionFilesDict.has_key(keyid):
                notify().sent(keyid, filename, transferID, 'failed', None, 'not found outgoing queue to that user')
                return
            if len(_InActionFilesDict[keyid]) == 0:
                notify().sent(keyid, filename, transferID, 'failed', None, 'no files to that user found in the outgoing queue')
                return
            filename, transferID, description = _InActionFilesDict[keyid].pop(0)
            if err < 0 :
                try:
                    if sock is not None:
                        sock.close()
                except:
                    logger().exception()
                notify().sent(keyid, filename, transferID, 'failed', err, 'error connecting')
                return
            if not _SendingStreams.has_key(keyid):
                _SendingStreams[keyid] = []
            stream = TCPMessageStream(sock, self.reactor)
            newSender = FileSender(self.reactor, stream, keyid)
            _SendingStreams[keyid].append(newSender)
            newSender.newFile(filename, transferID, description)
        def onConnect( connector ) :
            global _InActionFilesDict 
            if connector.getError() != 0:
                if not _InActionFilesDict.has_key(keyid):
                    notify().sent(keyid, filename, transferID, 'failed', None, 'not found outgoing queue to that user')
                    return
                if len(_InActionFilesDict[keyid]) == 0:
                    notify().sent(keyid, filename, transferID, 'failed', None, 'no files to that user found in the outgoing queue')
                    return
                filename, transferID, description = _InActionFilesDict[keyid].pop(0)
                notify().sent(keyid, filename, transferID, 'failed', None, 'connection error: ' + connector.getErrorMsg())
                return
            CSpaceConnector(connector.sock, keyid, 'receive', self.reactor, onConnectUser)
        tcpConnect(('127.0.0.1', self.appletport), self.reactor, onConnect)
        
    def onService(self, sslConn, peerKey, contactName, contactKeyID, incomingName):
        connectionId = self.service.appletServer.incoming.addIncoming(sslConn, peerKey)
        def onAccept(err, sock, keyid):
            global _ReceivingStreams
            if err < 0 :
                try:
                    if sock is not None:
                        sock.close()
                except:
                    logger().exception('onService.onAccept err=%d, keyid=%s' % (err, keyid))
                notify().receive('', 'failed', keyid, err, 'error: ' + str(err))
                return
            if not _ReceivingStreams.has_key(keyid):
                _ReceivingStreams[keyid] = []
            stream = TCPMessageStream(sock, self.reactor)
            newReceiver = FileReceiver(self.reactor, keyid, stream)
            _ReceivingStreams[keyid].append(newReceiver)
        def onConnect(connector, keyid):
            if connector.getError() != 0 :
                notify().receive('', 'failed', keyid, None, 'error connecting to CSpace on local machine')
                return
            sock = connector.getSock()
            CSpaceAcceptor(sock, connectionId, self.reactor, lambda e,s: onAccept(e, s, keyid))
        # logger().info('onService contactName=%s' % contactName)
        tcpConnect(('127.0.0.1', self.appletport), self.reactor, lambda c: onConnect(c, str(contactKeyID)))
    
    #------------------------------------------------------------------------------
    
    def test_ssl0(self):
        from ncrypt0.rsa import RSAKey, RSAError
        # from cspace.network.sslutil import makeSSLContext
        from ncrypt0.ssl import SSLContext, SSLConnection, SSL_VERIFY_MODE_NONE, SSL_VERIFY_MODE_SELF_SIGNED, SSL_METHOD_TLSv1
        from socket import socket
        from OpenSSL.SSL import Context, Connection, WantReadError, TLSv1_METHOD, OP_NO_SSLv2, VERIFY_PEER, VERIFY_FAIL_IF_NO_PEER_CERT
        from OpenSSL.crypto import X509, FILETYPE_PEM, load_privatekey, load_certificate
        
        def b(s):
            return s
        
        def makeCertifiace(userName, rsaKey):
            key = load_privatekey(FILETYPE_PEM, rsaKey.toPEM_PrivateKey())
            cert = X509()
            cert.get_subject().CN = userName
            cert.set_issuer(cert.get_subject())
            cert.set_pubkey(key)
            cert.set_notBefore("20000101000000Z")
            cert.set_notAfter("22000101000000Z")
            cert.set_serial_number(1)
            cert.sign(key, "sha1")
            return cert

        def makeSSLContext(userName, rsaKey) :
            cert = makeCertifiace(userName, rsaKey)
            sslContext = SSLContext( SSL_METHOD_TLSv1 )
            # store = sslContext.c.get_cert_store()
            # store.add_cert(cert)
            # sslContext.c.set_verify(VERIFY_PEER | VERIFY_FAIL_IF_NO_PEER_CERT, lambda connection, x509, errno, depth, preverifyOK: preverifyOK)
            # sslContext.c.set_options(OP_NO_SSLv2)
            sslContext.setCertificate( cert )
            sslContext.setPrivateKey( rsaKey )
            sslContext.checkPrivateKey()
            sslContext.setVerifyMode(SSL_VERIFY_MODE_NONE)
            return sslContext
            
        def socket_pair():
            port = socket()
            port.bind(('', 0))
            port.listen(1)
            client = socket()
            client.setblocking(False)
            client.connect_ex(("127.0.0.1", port.getsockname()[1]))
            client.setblocking(True)
            server = port.accept()[0]
            server.send(b("x"))
            print client.recv(1024)
            client.send(b("y"))
            print server.recv(1024)
            server.setblocking(False)
            client.setblocking(False)
            return (server, client)

        def clientContext():
            ps = self.profile('CSpace_cli', 'CSpaceProfiles_cli')
            # entry = self.settings('CSpace_cli').getString('Settings/SavedProfile')
            userName = ps.name
            keyId = ps.keyId
            rsaKey = ps.rsaKey
            print 'clientContext', userName
            cont = makeSSLContext(userName, rsaKey)
            return cont
        
        def serverContext():         
            ps = self.profile('CSpace_srv', 'CSpaceProfiles_srv')
            # entry = self.settings('CSpace_srv').getString('Settings/SavedProfile')
            userName = ps.name
            keyId = ps.keyId
            rsaKey = ps.rsaKey
            print 'serverContext', userName
            cont = makeSSLContext(userName, rsaKey)
            return cont
       
        def clientFactory(context, socket):
            client = Connection(context, socket)
            client.set_connect_state()
            return client
        
        def serverFactory(context, socket):
            server = Connection(context, socket)
            server.set_accept_state()
            return server
        
        def handshake(client, server):
            conns = [client, server]
            while conns:
                for conn in conns:
                    try:
                        conn.do_handshake()
                    except WantReadError:
                        pass
                    else:
                        conns.remove(conn)
                        
        srvSock, cliSock = socket_pair()
        srvContext, cliContext = (serverContext(), clientContext())
        server, client = (serverFactory(srvContext.c, srvSock), clientFactory(cliContext.c, cliSock))
        handshake(client, server)
        server.setblocking(True)
        client.setblocking(True)
        count = server.send(b('xy'))
        print client.recv(2)
        count = client.send(b('yx'))
        print server.recv(2)
        print 'srv', server.get_peer_certificate()
        print 'cli', client.get_peer_certificate().get_subject().commonName

    def test_ssl1(self):
        dumped = '''-----BEGIN RSA PRIVATE KEY-----
MIIBDQIBAAKCAQEA0I3bi9M/Ee7DkrIIPUT41Qf5jVHmbnks5r5Oe96B9EzkTgFO
Wwi8/SkT03zHG+0X0+2eIlgHtT2eKMFfasRtmimAEHP3jI89vBY+BA64vnqmgQVo
CxVNnLkNcMICTdgq/qHpIKBG+6kUW0++KCierRJ0xqc7Xgs0RoN1DnznY2mQm69D
J8uynu9JAW2wIl+Iqz6DaXdnfXBHFEdB3/RTUUzjqL+ucr9HCqCYHeOcG/xszSxF
1q+nY8F+IovSI6Q+XXFiHwqdYXixeMtHNcSa5cJkpz98FXQJF7oDcLuo7MM9/GYW
94co66dHoQeNwoncupBFu+CEKJUBe95m1n1YnQIDAQAB
-----END RSA PRIVATE KEY-----'''
        k = RSAKey()
        k.fromPEM_PrivateKey(dumped)
        print k.toPEM_PublicKey()
        

    def test_ssl2(self):
        from nitro.ssl import sslConnect, sslAccept
        from nitro.tcp import tcpListen, tcpConnect
        from ncrypt0.ssl import SSLContext, SSLConnection
        from ncrypt0.rsa import RSAKey, RSAError
        from cspace.network.sslutil import makeClientSSLContext, makeServerSSLContext

        reactor = SelectReactor()
        
        def clientContextCertVerifyCallback(conn, cert, errno, depth, preverify_ok):
            print 'clientContextCertVerifyCallback', conn, errno, depth, preverify_ok, cert.get_pubkey(), cert.get_subject().CN
            return preverify_ok
        
        def clientContext():
            # ps = self.profile('CSpace_cli', 'CSpaceProfiles_cli')
            # entry = self.settings('CSpace_cli').getString('Settings/SavedProfile')
            # userName = ps.name
            # keyId = ps.keyId
            # rsaKey = ps.rsaKey
            rsaKey = RSAKey()
            rsaKey.fromPEM_PrivateKey(open('cs/CliPrivateKey', 'rb').read(), open('cs/CliSavedPassword', 'rb').read().strip('"'))
            userName = open('cs/CliName', 'rb').read()
            cont = makeClientSSLContext(userName, rsaKey, clientContextCertVerifyCallback)
            print 'clientContext', userName
            print rsaKey.toPEM_PublicKey()
            return cont
        
        def serverContextCertVerifyCallback(conn, cert, errno, depth, preverify_ok):
            print 'serverContextCertVerifyCallback', conn, errno, depth, preverify_ok, cert.get_pubkey(), cert.get_subject().CN
            return preverify_ok

        def serverContext():         
#            ps = self.profile('CSpace_srv', 'CSpaceProfiles_srv')
#            entry = self.settings('CSpace_srv').getString('Settings/SavedProfile')
#            userName = ps.name
#            keyId = ps.keyId
#            rsaKey = ps.rsaKey
            rsaKey = RSAKey()
            rsaKey.fromPEM_PrivateKey(open('cs/SrvPrivateKey', 'rb').read(), open('cs/SrvSavedPassword', 'rb').read().strip('"'))
            userName = open('cs/SrvName', 'rb').read()
            cont = makeServerSSLContext(userName, rsaKey, serverContextCertVerifyCallback)
            print 'serverContext', userName
            print rsaKey.toPEM_PublicKey()
            return cont

        def onConnect(connector, sslContext):
            print 'onConnect', connector
            # connector.getSock().write('xy')
            sslConn = SSLConnection( sslContext, connector.getSock() )
            print sslConn.getPeerCertificate()
            sslConn.conn.set_connect_state()
            connectOp = sslConnect( sslConn, reactor, lambda err: onSSLConnect(err, sslConn) )
            
        def onWrite(connector):
            print 'onWrite', connector

        def onSSLAccept(err, sslConn):
            peerCert = sslConn.conn.get_peer_certificate()
            peerKey = RSAKey()
            peerKey.fromPKey_PublicKey(peerCert.get_pubkey())
            print 'onSSLAccept', peerCert.get_subject().CN
            
        def onSSLConnect(err, sslConn):
            print 'onSSLConnect', err, sslConn.conn.get_peer_certificate().get_subject().CN
        
        def onIncomingTCP(sock, sslContext):
            print 'onIncomingTCP', sock
            # print sock.recv(2)
            sslConn = SSLConnection( sslContext, sock )
            # print sslConn.getPeerCertificate()
            sslConn.conn.set_accept_state()
            acceptOp = sslAccept( sslConn, reactor, lambda err: onSSLAccept(err, sslConn) )
            
        def onRead(sock):
            print 'onRead', sock
            
        if sys.argv.count('server'):
            listener = tcpListen( ('',0), reactor, None )
            localPort = listener.getSock().getsockname()[1]
            srvContext = serverContext()
            listener.setCallback( lambda sock: onIncomingTCP(sock, srvContext) )
            listener.enable( True )
            print 'server port is', localPort
        elif sys.argv.count('client'):
            localPort = int(sys.argv[sys.argv.index('client')+1])
            cliContext = clientContext()
            connector = tcpConnect(('127.0.0.1', localPort), reactor, lambda conn: onConnect(conn, cliContext))
        else:        
            listener = tcpListen( ('',0), reactor, None )
            localPort = listener.getSock().getsockname()[1]
            srvContext, cliContext = (serverContext(), clientContext())
            listener.setCallback( lambda sock: onIncomingTCP(sock, srvContext) )
            listener.enable( True )
            connector = tcpConnect(('127.0.0.1', localPort), reactor, lambda conn: onConnect(conn, cliContext))
        reactor.run()

#------------------------------------------------------------------------------ 

def controlCommand(command, data=''):
    def _write():
    # logger().info('controlCommand %s [%s]' % (command, str(data)))
    # request = HttpRequest(reactor)
    # postData = urllib.urlencode(dict(command=command, data=data), True)
    # return request.post(url, postData)
        try:
            sys.stdout.write('[%s] %s\n' % (command, str(data)))
            sys.stdout.flush()
        except:
            logger().exception('write command error')
    reactor.callLater(0, _write)

#------------------------------------------------------------------------------ 

class Notifier():
    def sent(self, keyID, filename, transferID, status, err=None, msg=''):
        controlCommand('sent', (str(keyID), filename, transferID, status, err, '"%s"' % msg))

    def receive(self, filename, status, peer, err=None, msg=''):
        controlCommand('receive', (str(peer), filename, status, err, '"%s"' % msg))

    def contact_status(self, name, status):
        controlCommand('contact', (name, status))

def notify():
    global _Notifier
    if _Notifier is None:
        _Notifier = Notifier()
    return _Notifier

#------------------------------------------------------------------------------ 

class CSpace_stdout:
    softspace = 0
    def read(self): pass
    def write(self, s):
        # sys.stdout.write(time.strftime('%H:%M:%S') + '    ' + s)
        sys.stderr.write(s)
        sys.stderr.flush()
    def flush(self): pass
    def close(self): pass
    
#------------------------------------------------------------------------------ 
    
def popFileName(keyid):
    global _OutgoingFilesDict
    SendToKeyID = str(keyid)
    if not _OutgoingFilesDict.has_key(SendToKeyID):
        logger().critical('popFileName no files for %s' % SendToKeyID)
        return '', -1, ''
    try:
        filename, transferID, description = _OutgoingFilesDict[SendToKeyID].pop(0)
        # logger().info('popFileName %s %s %s %s' % (SendToKeyID, filename, str(transferID), description))
        return filename, transferID, description
    except:
        logger().exception('exception in popFileName(%s)' % str(SendToKeyID))
        logger().info(str(_OutgoingFilesDict.keys()))
    return '', -1, ''

def pushFileName(keyid, filename, transferID, description):
    # global _OutgoingQueue
    global _OutgoingFilesDict
    SendToKeyID = str(keyid)
    if not _OutgoingFilesDict.has_key(SendToKeyID):
        _OutgoingFilesDict[SendToKeyID] = []
    _OutgoingFilesDict[SendToKeyID].append((filename, transferID, description))
    # _OutgoingQueue.append(SendToKeyID)
    # logger().info('pushFileName %s %s %s %s' % (SendToKeyID, filename, str(transferID), description))
    
#------------------------------------------------------------------------------ 

class CSpaceConnector( object ) :
    def __init__( self, sock, remoteUser, remoteService, reactor, callback ) :
        self.stream = TCPStream( sock, reactor )
        self.stream.setCloseCallback( self._onClose )
        self.stream.setErrorCallback( self._onError )
        self.stream.setInputCallback( self._onInput )
        self.stream.initiateRead( 1 )
        self.stream.writeData( 'CONNECT %s %s\r\n' % (remoteUser,remoteService) )
        self.response = ''
        self.op = AsyncOp( callback, self.stream.close )

    def getOp( self ) : return self.op

    def _onClose( self ) :
        self.stream.close()
        self.op.notify( -1, None )

    def _onError( self, err, errMsg ) :
        self._onClose()

    def _onInput( self, data ) :
        self.response += data
        if self.response.endswith('\n') :
            if not self.response.startswith('OK') :
                self._onClose()
                return
            self.stream.shutdown()
            sock = self.stream.getSock()
            self.op.notify( 0, sock )

class CSpaceAcceptor( object ) :
    def __init__( self, sock, connectionId, reactor, callback ) :
        self.stream = TCPStream( sock, reactor )
        self.stream.setCloseCallback( self._onClose )
        self.stream.setErrorCallback( self._onError )
        self.stream.setInputCallback( self._onInput )
        self.stream.initiateRead( 1 )
        self.stream.writeData( 'ACCEPT %s\r\n' % connectionId )
        self.response = ''
        self.op = AsyncOp( callback, self.stream.close )

    def getOp( self ) : return self.op

    def _onClose( self ) :
        self.stream.close()
        self.op.notify( -1, None )

    def _onError( self, err, errMsg ) :
        self._onClose()

    def _onInput( self, data ) :
        self.response += data
        if self.response.endswith('\n') :
            if not self.response.startswith('OK') :
                self._onClose()
                return
            self.stream.shutdown()
            sock = self.stream.getSock()
            self.op.notify( 0, sock )

#------------------------------------------------------------------------------ 
# need to send files inside same stream
# close stream if only connection error or connection lost
class FileSender(object):
    CHUNK_SIZE = (2 ** 16) - 256

    def __init__(self, reactor, stream, peer):
        self.reactor = reactor
        self.stream = stream
        self.file = None
        self.filename = None
        self.transferID = None
        self.description = None
        self.peer = peer
        self.waitOp = None
        self.currentOp = None
        self.bytesSent = 0
        self.length = 0
        self.totalBytesSent = 0
        self.lastPacketTime = 0
        self.state = 'ready'
        self.ready = time.time()
        self.stream.setInputCallback(self.dataReceived)
        self.stream.setCloseCallback(self.connectionLost)
        self.stream.setErrorCallback(self.connectionError)
        self.stream.enableRead(True)
        # self.doWaitNewFiles()
        # logger().info('FileSender.__init__ peer=%s, %s' % (self.peer, str(self.stream.getSock().getsockname())))

    # def __del__(self):
        # logger().info('FileSender.__dell__ peer=%s totalSent=%d state=%s' % (self.peer, self.totalBytesSent, self.state))

    def newFile(self, filename, transferID, description):
        self.filename, self.transferID, self.description = (filename, transferID, description)
        try:
            self.file = file(self.filename, 'rb')
            self.file.seek(0, 2)
            self.length = self.file.tell()
            self.file.seek(0)
        except:
            self.state = 'error'
            self.failed('failed', 'error reading input file')
            self.close_stream()
            return
        try:
            self.stream.sendMessage(struct.pack("!Q", self.length))
        except:
            self.state = 'lost'
            self.failed('failed', 'error writing data, seems connection were lost')
            self.close_stream()
            return
        self.bytesSent = 0
        self.state = 'sending'
        self.doSend()

    def doSend(self):
        if self.file is None:
            logger().critical('file is None: ' + str((self.peer, self.state, self.transferID, self.filename, self. self.bytesSent, self.totalBytesSent)))
            return
        chunk = self.file.read(self.CHUNK_SIZE)
        if not chunk:
            self.state = 'waiting'
            self.doWaitFinishSending()
            return
        self.stream.sendMessage(chunk)
        self.bytesSent += len(chunk)
        self.currentOp = self.reactor.callLater(0, self.doSend)

    def doWaitFinishSending(self):
        #all data was sent
        #now need to wait to finish receiving
        #but we do not want to wait too long
        #if we did not receive any response
        #we assume the connection was lost
        def sent_timeout():
            if time.time() - self.lastPacketTime > 30:
                del self.waitOp
                self.waitOp = None
                self.state = 'ready'
                self.ready = time.time()
                self.failed('timeout', 'other side not responding')
                # self.doWaitNewFiles()
            else:
                self.doWaitFinishSending()
        self.waitOp = self.reactor.callLater(60, sent_timeout)

    def dataReceived(self, data):
        self.lastPacketTime = time.time()
        bytesReceived ,= struct.unpack("!Q", data[:8])
        if bytesReceived == self.length:
            self.state = 'ready'
            self.ready = time.time()
            self.done()
            # self.doWaitNewFiles()
            return
        if bytesReceived > self.length:
            self.state = 'ready'
            self.ready = time.time()
            self.failed('failed', 'data did not sent correctly')
            # self.doWaitNewFiles()

    def connectionLost(self):
        #print 'FileSender.connectionLost'
        st = self.state 
        self.state = 'lost'
        if st in ['sending', 'waiting']:
            self.failed('failed', 'connection lost while sending')
        self.close_stream()

    def connectionError(self, err, errMsg):
#        print 'FileSender.connectionError', errMsg
        st = self.state
        self.state = 'error'
        if st in ['sending', 'waiting']:
            self.failed('failed', 'connection error while sending: ' + errMsg)
        self.close_stream()

    def done(self):
        self.totalBytesSent += self.bytesSent
        self.close_file()
        if self.transferID is None or self.filename is None:
            logger().critical('sending finished with some error: ' + str((self.peer, self.state, self.transferID, self.filename, self. self.bytesSent, self.totalBytesSent)))
        notify().sent(self.peer, self.filename, self.transferID, 'finished')
        self.close_wait_op()
        self.clear_state()

    def failed(self, status, msg):
        self.close_file()
        notify().sent(self.peer, self.filename, self.transferID, status, None, msg)
        self.close_wait_op()
        self.clear_state()

    def close_wait_op(self):
        if self.waitOp is not None:
            self.waitOp.cancel()
            self.waitOp = None

    def close_file(self):
        try:
            self.file.close()
        except:
            pass
        
    def clear_state(self):
        self.file = None
        self.filename = None
        self.transferID = None
        self.description = None

    def close_stream(self):
        global _SendingStreams
        if self.state in ['sending', 'waiting']:
            self.failed('failed', 'closing sending stream')
        # self.close_wait_op()
        # self.close_file()
        # self.clear_state()
        if not self.stream.stream.shutdownFlag:
            self.stream.close()
        if _SendingStreams.has_key(self.peer) and self in _SendingStreams[self.peer]:
            _SendingStreams[self.peer].remove(self)
        # logger().info('FileSender.close_stream with %s, more streams: %d' % (self.peer, len(_SendingStreams[self.peer])))


class FileReceiver(object):
    def __init__(self, reactor, peer, stream):
        self.reactor = reactor
        self.peer = peer
        self.stream = stream
        self.bytesReceived = 0
        self.totalBytesReceived = 0
        self.length = 0
        self.fd = 0
        self.filename = None
        self.gotLength = False
        self.state = 'ready'
        self.ready = time.time()
        self.stream.setInputCallback(self.dataReceived)
        self.stream.setCloseCallback(self.connectionLost)
        self.stream.setErrorCallback(self.connectionError)
        self.stream.enableRead(True)
        # logger().info('FileReceiver.__init__ peer=%s, %s' % (self.peer, str(self.stream.getSock().getsockname())))

    # def __del__(self):
        # logger().info('FileReceiver.__dell__ peer=%s totalReceived=%d state=%s' % (self.peer, self.totalBytesReceived, self.state))

    def openFile(self):
        global _TempLocation
        self.fd, self.filename = tempfile.mkstemp('', '', _TempLocation)

    def closeFile(self):
        os.fsync(self.fd)
        os.close(self.fd)
        
    def clearFile(self):
        self.fd = None
        self.filename = None

    def dataReceived(self, data):
        if self.state == 'ready':
            self.openFile()
            self.length ,= struct.unpack("!Q", data[:8])
            self.bytesReceived = 0
            data = data[8:]
            self.state = 'receiving'
        os.write(self.fd, data)
        self.bytesReceived += len(data)
        self.stream.sendMessage(struct.pack("!Q", self.bytesReceived))
        if self.bytesReceived == self.length:
            self.closeFile()
            self.totalBytesReceived += self.bytesReceived
            self.state = 'ready'
            self.ready = time.time()
            notify().receive(self.filename, 'finished', self.peer)
            self.clearFile()

    def connectionLost(self):
        if self.state == 'receiving':
            self.closeFile()
            notify().receive(self.filename, 'failed', self.peer, None, 'connection lost')
            self.clearFile()
        self.close_stream()
        self.state = 'lost'

    def connectionError(self, err, errMsg):
        if self.state == 'receiving':
            self.closeFile()
            notify().receive(self.filename, 'failed', self.peer, err, 'connection error: ' + errMsg)
            self.clearFile()
        self.close_stream()
        self.state = 'error'

    def close_stream(self):
        global _ReceivingStreams
        if not self.stream.stream.shutdownFlag:
            self.stream.close()
        if _ReceivingStreams.has_key(self.peer) and self in _ReceivingStreams[self.peer]:
            _ReceivingStreams[self.peer].remove(self)
        # logger().info('FileReceiver.close_stream with %s, more streams: %d' % (self.peer, len(_ReceivingStreams[self.peer])))

#------------------------------------------------------------------------------ 

def CA_KEY():
    return '''-----BEGIN RSA PRIVATE KEY-----
MIIEpgIBAAKCAQEA1XegQ7Qr1lbIf0ldWLkVEve1rqahz1A6AjFGTkXXHagTlT5y
OncQoIZXDvswJGjawjgG2J1DetwLfK+Xw746QVp/SfQAl+F+hX29BppCctPL7avB
NmTvq2VTR2QpAdRo9tv3toy/LnuUWFxere1HWmTq6+tiNuVQALy56yex5HV1FSGT
gnEKbQCe8jYRoCf3mi9fx56Ynd9ZDY8Usxl3ZGS2W+8dbKrC8srBtN65DIWsmvQS
Xw3EOHF/zOGzGkYzp8BgxjOC6pgdSGjcW6NzrdnkCcps2g002gDOdHKrMs5qDznS
3xVm7BsGAfAjvywsCPY45AkUIFOTwBl6AKTIrQIDAQABAoIBAQDMhrl/JPl7e9rO
WjSd9XdDnSLuG6mdQHjT8PIzvKbHO2rH6/T2H25G33A6YmFWAUDYJWYp1UP6SyqW
ZIc8fN7EDzk2WhrXaq4WqMqbsOFJs7QIHDAbNcqMpaCNHmJL5oBLRaapuWDT4IPZ
xWbRri5XZanQMM8BWeS1UB2yOW4wPLnV66QAO85dWGPjrCu2blHdEOHUsXgY/OKC
rFJEwF9gcFumC67K8WOk53cVRb1WbIiSzI29IhprFnk8aNho7NAra1SFDL6VtDkZ
V9SFQE3U8Dn0yheb/KYuiQGSA25CffbBvzZuoZI2NkqB51tjZpfmGGfFCuaEN8xp
OdxxrDfhAoGBAO1jRp1hxTglTvBMEDzV2H1CHoCTc6wO5+u+klncNeV/di3VLLdR
YXy0b809s7Jiy5+BwWPD+3jNgVbV/JkB01GS3ndCkpV5vyuuLNC7phwlg7fiXBge
t408iV280BvNYVe0oMfiiPqN3ZXtfnhO3mTpU2zT+7NcpCF8ocIGhVWvAoGBAOY0
OvD9vZpGOz/3uB1hzT4W0s8RpZq/m9P7/DhwEuDadpbJMzqrDh2L1lsvdZMfGk3P
f5Zl6J0ZQ9WFnMYBVep+1hsF7EE4aTUgcayYTiae0RKYddCMany+xTpmZtb+9pL1
XgIGqMujpNvoA5IFFpQoq1GJRgym/Y7lWJSKxTpjAoGBANX0sGSRBmxAVBGIvOnX
47OhsFQ6kfr1xFpZ/RY4v+sFIsLUa+Ud2DvJdSsK/bc+DEDLdj4xGaobrwNRAsX1
Oz0+nnvm6K8IeCEqbwIC3whnV3yY7GBg7xbBBR4TW45zYdTm3DLMHqGU066Zy28r
xo41LhfcR/O0/8mexzxVHD/JAoGBAIUEly5boGx9uoza6jNoKP2AmK14J/YEU9mp
GHPQJshw0+eLOSPkZZKjE+i6wriRV1Sw2qugFUp5p93Ah/dOHEQUqEkTIhIJs5k0
NxshIr9kM7EIEcPA72NHpJV7SF9hEj+Wsox/JpgM5hz/sth0Qji4S04hAS4cbBVe
5tFmYlK7AoGBALr65+WrWsVWKM7rtzUNLD5DswXlwMX3peY3HrYt3Hyh/PZJeDGu
LTWy3ruVqCFbLwqOvIJNLJCPBY4/fY4NkgwCrsbxPpYmNmehgbQvP68XyxI9Jsuk
buF3X83VvHBpcuvJIvLq2/YLmH5SnxkR63aHBJRXEr2N2uRfGq7sv5Hi
-----END RSA PRIVATE KEY-----
'''

def CA_CERT():
    return '''-----BEGIN CERTIFICATE-----
MIIC7DCCAdSgAwIBAwIBATANBgkqhkiG9w0BAQUFADAUMRIwEAYDVQQDEwljYS5j
c3BhY2UwIhgPMjAwMDAxMDEwMDAwMDBaGA8yMjAwMDEwMTAwMDAwMFowFDESMBAG
A1UEAxMJY2EuY3NwYWNlMIIBIjANBgkqhkiG9w0BAQEFAAOCAQ8AMIIBCgKCAQEA
1XegQ7Qr1lbIf0ldWLkVEve1rqahz1A6AjFGTkXXHagTlT5yOncQoIZXDvswJGja
wjgG2J1DetwLfK+Xw746QVp/SfQAl+F+hX29BppCctPL7avBNmTvq2VTR2QpAdRo
9tv3toy/LnuUWFxere1HWmTq6+tiNuVQALy56yex5HV1FSGTgnEKbQCe8jYRoCf3
mi9fx56Ynd9ZDY8Usxl3ZGS2W+8dbKrC8srBtN65DIWsmvQSXw3EOHF/zOGzGkYz
p8BgxjOC6pgdSGjcW6NzrdnkCcps2g002gDOdHKrMs5qDznS3xVm7BsGAfAjvyws
CPY45AkUIFOTwBl6AKTIrQIDAQABo0UwQzASBgNVHRMBAf8ECDAGAQH/AgEAMA4G
A1UdDwEB/wQEAwIBBjAdBgNVHQ4EFgQUl08vCVH5xDXQhHMBMaGDsO3GvMIwDQYJ
KoZIhvcNAQEFBQADggEBAKAuJUiwLp69665CBp0dag/wnA1qnVYzPUISc7FLaaZd
GjjRx3+e9I4pMWyNAA+50Y6wDKe3Qqva3xYwiZafysrj6kZaTwpH3FGuLhUlH6HH
1kiqxjUAVNPtxCaq+5bzLip8GurZPjW6LLSq+4sg1SGxWCnnYvURwlVsxpYIRh66
q6hJvlJXK8eXnzI7TV9s9KFV99iZ/zkfY0YpqMhLdyQiTCcVCIpefg9xnEi13fng
N8mjgjSWmdOeQ2wmPmWl2ueprNCeBMUxMWm4d/cIK00RiD4Z11mORMkE8qbb7Rs8
hIjsk+ffqMW/LsjgJkdsz9MhrNE7ffDyGMmZbqbBoqc=
-----END CERTIFICATE-----
'''

#------------------------------------------------------------------------------ 

class AppLogger():
    def __init__(self, write_method=sys.stdout.write):
        self._w = write_method
    def write(self, s):
        self._w(s)
    def info(self, s=''):
        self.write(s+'\n')
    def critical(self, s=''):
        self.write(s+'\n')
    def debug(self, s=''): 
        self.write(s+'\n')
    def exception(self, s=''):
        self.write(traceback.format_exc()+'\n')

def logger():
    global _AppLogger
    if _AppLogger is None:
        def _w(s):
            sys.stderr.write(s)
            sys.stderr.flush()
        _AppLogger = AppLogger(_w)
    return _AppLogger

def initLog(logLevel=0, logFilename='cspace.log', logWriteMode='wb', use_logging_module=True, debug=False):
    global _AppLogger
    if _AppLogger is not None:
        logger().critical('CSpaceApplication.initLog WARNING AppLogger already initialized')
        return
    if use_logging_module:
        if debug:
            logging.basicConfig(level=logLevel, stream=CSpace_stdout(), format='%(name)30s    %(levelname)5s   %(message)s', )
        else:
            logging.basicConfig(level=logLevel, filename=logFilename, filemode=logWriteMode, format='%(name)30s    %(levelname)5s   %(message)s', )
        _AppLogger = logging.getLogger('dhn_cspace')
    else:
        if not debug:
            sys.stderr = open(logFilename, logWriteMode)
#        def _w(s):
#            sys.stderr.write(s)
#            sys.stderr.flush()
#        _AppLogger = AppLogger(_w)

#------------------------------------------------------------------------------ 

def parseCommandLine():
    oparser = optparse.OptionParser()
    oparser.add_option("-r", "--register", dest="register", action="store_true", help="generate Private Key and register new key ID")
    oparser.add_option("-s", "--seednodes", dest="seednodes", action="append", help="seed nodes")
    oparser.add_option("-l", "--logfile", dest="logfile", help="where to put the output messages, default is cspace.log")
    oparser.add_option("-x", "--xmlrpc", dest="xmlport", type="int", help="specify port for XMLRPC control")
    oparser.add_option("-H", "--homedir", dest="homedir", help="home directory location")
    oparser.add_option("-t", "--tmpdir", dest="tmpdir", help="temporary files location")
    oparser.add_option("-d", "--debug", dest="debug", action="store_true", help="redirect output to stderr")
    oparser.set_default('debug', False)
    oparser.set_default('register', False)
    oparser.set_default('xmlport', 0)
    oparser.set_default('logfile', 'cspace.log')
    oparser.set_default('homedir', '')
    oparser.set_default('tmpdir', '')
    oparser.set_default("seednodes", ['208.78.96.185:10001'])  # ('210.210.1.102', 10001),
    (options, args) = oparser.parse_args()
    options.xmlport = int(options.xmlport)
    return options, args

#------------------------------------------------------------------------------ 

def test():
    (options, args) = parseCommandLine()
    initLog(logLevel=30, debug=options.debug)
    reactor = SelectReactor()
    a = Application(reactor, options, args)
    a.test_ssl2()
    # a.reactor.run()
    
def generateCA():
    import OpenSSL
    key = OpenSSL.crypto.PKey()
    key.generate_key(OpenSSL.crypto.TYPE_RSA, 2048)
    
    ca = OpenSSL.crypto.X509()
    ca.set_version(3)
    ca.set_serial_number(1)
    ca.get_subject().CN = "ca.cspace"
    ca.set_notBefore("20000101000000Z")
    ca.set_notAfter("22000101000000Z")
    ca.set_issuer(ca.get_subject())
    ca.set_pubkey(key)
    ca.add_extensions([
      OpenSSL.crypto.X509Extension("basicConstraints", True,
                                   "CA:TRUE, pathlen:0"),
      OpenSSL.crypto.X509Extension("keyUsage", True,
                                   "keyCertSign, cRLSign"),
      OpenSSL.crypto.X509Extension("subjectKeyIdentifier", False, "hash",
                                   subject=ca),
      ])
    ca.sign(key, "sha1")
    cert_str = OpenSSL.crypto.dump_certificate(OpenSSL.crypto.FILETYPE_PEM, ca)
    key_str = OpenSSL.crypto.dump_privatekey(OpenSSL.crypto.FILETYPE_PEM, key)
    fout = open('ca.pem', 'wb')
    fout.write(cert_str.decode('ascii'))
    fout.close()
    fout = open('ca.key', 'wb')
    fout.write(key_str.decode('ascii'))
    fout.close()
    fin = open('ca.pem', 'rb')
    ca_cert_ = OpenSSL.crypto.load_certificate(OpenSSL.crypto.FILETYPE_PEM, fin.read())
    fin.close()

def main():
    global reactor
    (options, args) = parseCommandLine()
    initLog(logFilename=options.logfile.strip('"'), debug=options.debug)
    reactor = SelectReactor()
    a = Application(reactor, options, args)
    if a.request_registration:
        try:
            a.register(a.username, a.password)
        except:
            logger().exception('ERROR during registration')
            return -1
        return 0
    try:
        a.init()
    except:
        logger().exception('ERROR during initialization')
        return -1
    try:
        a.run()
    except:
        logger().exception('ERROR:')
        return -1
    return 0

#------------------------------------------------------------------------------ 
        
if __name__ == "__main__":
    main()
    # test()
    # generateCA()
    
        



#!/usr/bin/python
#transport_udp_session.py
#
#
#    Copyright DataHaven.NET LTD. of Anguilla, 2006-2013
#    Use of this software constitutes acceptance of the Terms of Use
#      http://datahaven.net/terms_of_use.html
#    All rights reserved.
#
#


import os
import time
import struct
import cStringIO
import random


from twisted.internet import reactor
from twisted.internet.task import LoopingCall 


import dhnio
import settings
import misc
import identity
import identitycache
import tmpfile 
import nameurl

import transport_udp
import automat

#------------------------------------------------------------------------------ 

CODES = {'p:':  'PING',
         'g:':  'GREETING',
         'd:':  'DATA',
         'r:':  'REPORT',
         'a:':  'ALIVE',}

COMMANDS = {'PING':         'p:',
            'GREETING':     'g:',
            'DATA':         'd:',
            'REPORT':       'r:',
            'ALIVE':        'a:',}

BLOCK_SIZE_LEVELS = {   #0: 508,
                        0: 1460,
                        1: 2048,
                        2: 3072,
                        3: 4096,
                        4: 5120,
                        5: 6144,
                        6: 7168,
                        7: 8124,
                        }
MAX_BLOCK_SIZE_LEVEL = len(BLOCK_SIZE_LEVELS) - 1

#BLOCK_SIZE = 4096
TIME_OUT = 20
BLOCK_RETRIES = 5
MAX_WINDOW_SIZE = 32
MIN_WINDOW_SIZE = 1

#------------------------------------------------------------------------------ 

_ParentObject = None

_IdToAddress = {}
_AddressToId = {}

_SendStatusCallback = None
_ReceiveStatusCallback = None
_SendControlFunc = None
_ReceiveControlFunc = None
_RegisterTransferFunc = None
_UnRegisterTransferFunc = None
_StateChangedCallbackFunc = None

_TimerPing = None
_TimerAlive = None
_Timer30sec = None

_SendingDelay = 0.01
_SendingTask = None

_OutboxQueueDelay = 0.01
_OutboxQueueTask = None

#------------------------------------------------------------------------------ 

def init(parent):
    global _ParentObject
    dhnio.Dprint(4, 'transport_udp_session.init')
    _ParentObject = parent
    reactor.callLater(0, process_outbox_queue)
    reactor.callLater(1, process_sending)
    StartTimers()

   
def shutdown():
    dhnio.Dprint(4, 'transport_udp_session.shutdown')
    StopTimers()
    
   
def A(address, event=None, arg=None):
    s = open_session(address)
    if event is not None:
        s.automat(event, arg)
    return s
      
      
def parent():
    global _ParentObject
    return _ParentObject

#------------------------------------------------------------------------------ 
    
class OutboxFile():
    def __init__(self, peer, remote_idurl, file_id, filename, filesize, transfer_id, description=''):
        self.peer = peer
        self.remote_idurl = remote_idurl
        self.file_id = file_id
        self.filename = filename
        self.transfer_id = transfer_id
        self.description = description
        self.size = filesize
        self.block_id = -1
        self.num_blocks = 0
        self.blocks = {} 
        self.bytes_sent = 0
        self.failed_blocks = {}
        self.has_failed_blocks = False
        self.started = time.time()
        self.timeout = max( 2*int(self.size/settings.SendingSpeedLimit()), 60)        
        # dhnio.Dprint(6, 'transport_udp_session.OutboxFile [%d] to %s' % (self.file_id, nameurl.GetName(self.remote_idurl)))

    def close(self):
        global _UnRegisterTransferFunc
        if _UnRegisterTransferFunc and self.transfer_id:
            _UnRegisterTransferFunc(self.transfer_id)
        # dhnio.Dprint(6, 'transport_udp_session.OutboxFile close [%d] to %s (%s)' % (self.file_id, nameurl.GetName(self.remote_idurl), str(self.peer)))

    def get_bytes_sent(self):
        return self.bytes_sent 

    def report_block(self, block_id):
        if not self.blocks.has_key(block_id):
            # dhnio.Dprint(2, 'transport_udp_session.report_block WARNING unknown block_id in REPORT packet received from %s: [%d]' % (self.remote_address, block_id))
            return
        self.bytes_sent += len(self.blocks[block_id])
        del self.blocks[block_id]

    def send(self):
        do_send = False
        while len(self.peer.sliding_window) < self.peer.sliding_window_size and self.block_id < self.num_blocks and self.block_id >= 0:
            if not self.send_block(self.block_id):
                break
            self.peer.sliding_window[(self.file_id, self.block_id)] = time.time()
            self.block_id += 1
        return do_send
    
    def send_block(self, block_id):
        global _SendControlFunc
        if not self.blocks.has_key(block_id):
            return False
        data = self.blocks[block_id]
        if _SendControlFunc is not None:
            more_bytes = _SendControlFunc(self.peer.last_bytes_sent, len(data))
            if more_bytes < len(data):
                return False
        self.peer.last_bytes_sent = len(data)
        datagram = COMMANDS['DATA']
        datagram += struct.pack('i', self.file_id)
        datagram += struct.pack('i', block_id)
        datagram += struct.pack('i', self.num_blocks)
        datagram += struct.pack('i', len(data))
        datagram += data
        return self.peer.sendDatagram(datagram)

    def mark_failed_block(self, block_id):
        if not self.failed_blocks.has_key(block_id):
            self.failed_blocks[block_id] = 0
        self.failed_blocks[block_id] += 1
        if self.failed_blocks[block_id] >= BLOCK_RETRIES:
            self.has_failed_blocks = True
            # dhnio.Dprint(8, 'transport_udp_session.mark_failed_block (%d,%d) is failed to send, so transfer %s is failed' % (self.file_id, block_id, self.transfer_id))
            return True
        return False
    
    def read_blocks(self):
        fin = open(self.filename, 'rb')
        block_id = 0
        while True:
            block_data = fin.read(BLOCK_SIZE_LEVELS[self.peer.block_size_level])
            if block_data == '':
                break
            self.blocks[block_id] = block_data
            block_id += 1  
        fin.close()
        self.num_blocks = block_id
        self.block_id = 0     
        
    def is_timed_out(self):
        return time.time() - self.started > self.timeout

    def cancel(self):
        self.timeout = 0

#------------------------------------------------------------------------------ 


class InboxFile():
    def __init__(self, peer, file_id, num_blocks):
        global _RegisterTransferFunc
        self.fd, self.filename = tmpfile.make("udp-in")
        self.peer = peer
        self.file_id = file_id
        self.num_blocks = num_blocks
        self.blocks = {}
        self.bytes_received = 0
        self.transfer_id = None
        if _RegisterTransferFunc is not None:
            self.transfer_id = _RegisterTransferFunc('receive', self.peer.remote_address, self.get_bytes_received, self.filename, -1, '')
        # dhnio.Dprint(6, 'transport_udp_session.InboxFile {%s} [%d] from %s' % (self.transfer_id, self.file_id, str(self.peer.remote_address)))

    def close(self):
        global _UnRegisterTransferFunc
        if _UnRegisterTransferFunc is not None and self.transfer_id is not None:
            _UnRegisterTransferFunc(self.transfer_id)
        # dhnio.Dprint(6, 'transport_udp_session.InboxFile.close {%s} [%d] from %s' % (self.transfer_id, self.file_id, str(self.peer.remote_address)))

    def get_bytes_received(self):
        return self.bytes_received

    def input_block(self, block_id, block_data):
        self.blocks[block_id] = block_data
        self.bytes_received += len(block_data) 
    
    def build(self):
        for block_id in range(len(self.blocks)):
            os.write(self.fd, self.blocks[block_id])
        os.close(self.fd)
        # dhnio.Dprint(10, 'transport_udp_server.InboxFile.build [%s] file_id=%d, blocks=%d' % (
        #     os.path.basename(filename), file_id, len(self.incommingFiles[file_id])))
    
#------------------------------------------------------------------------------ 

class TransportUDPSession(automat.Automat):
    fast = True
    timers = {'timer-60sec':     (60,    ['PING',]),
              'timer-3min':      (60*10,  ['PING',]),
              }
    def __init__(self, address, name, state, debug_level):
        self.remote_address = address
        self.remote_idurl = None
        self.remote_name = None
        self.last_alive_packet_time = 0
        self.incommingFiles = {}
        self.outgoingFiles = {}
        self.maxOutgoingFiles = 4
        self.receivedFiles = {}
        self.outbox_queue = []
        self.tries = 0
        self.sliding_window = {}
        self.sliding_window_size = MIN_WINDOW_SIZE
        self.last_bytes_sent = 0
        self.failedBlocks = {}
        self.hasFailedBlocks = False
        self.block_size_level = 0
        self.remote_id_request_last_time = 0
        automat.Automat.__init__(self, name, state, debug_level)
        
    def state_changed(self, oldstate, newstate):
        global _StateChangedCallbackFunc
        if _StateChangedCallbackFunc is not None:
            _StateChangedCallbackFunc(self.index, oldstate, newstate)
        
    def A(self, event, arg):
        #---CONNECTED---
        if self.state is 'CONNECTED':
            if event == 'datagram-received' :
                self.doReceiveData(arg)
            elif event == 'timer-30sec' and self.isAlive(arg) :
                self.doAlive(arg)
            elif event == 'shutdown' :
                self.state = 'CLOSED'
                self.doDestroyMe(arg)
            elif event == 'timer-30sec' and not self.isAlive(arg) :
                self.state = 'PING'
                self.doParentRequestRemoteID(arg)
                self.doPing(arg)
        #---AT_STARTUP---
        elif self.state is 'AT_STARTUP':
            if event == 'shutdown' :
                self.state = 'CLOSED'
                self.doDestroyMe(arg)
            elif event == 'init' :
                self.state = 'PING'
                self.doSaveIDURL(arg)
                self.doParentRequestRemoteID(arg)
                self.doPing(arg)
        #---PING---
        elif self.state is 'PING':
            if event == 'datagram-received' and self.isPingOrGreeting(arg) :
                self.state = 'GREETING'
                self.doReceiveData(arg)
                self.doGreeting(arg)
            elif event == 'datagram-received' and not self.isPingOrGreeting(arg) :
                self.doReceiveData(arg)
            elif event == 'shutdown' :
                self.state = 'CLOSED'
                self.doDestroyMe(arg)
            elif event == 'timer-60sec' :
                self.doParentRequestRemoteID(arg)
            elif event == 'timer-5sec' :
                self.doPing(arg)
            elif event == 'timer-3min' :
                self.state = 'CLOSED'
                self.doDestroyMe(arg)
                self.doParentRecreateMe(arg)
        #---GREETING---
        elif self.state is 'GREETING':
            if event == 'datagram-received' and self.isGreetingOrAlive(arg) :
                self.state = 'CONNECTED'
                self.doReceiveData(arg)
                self.doAlive(arg)
            elif event == 'datagram-received' and not self.isGreetingOrAlive(arg) :
                self.doReceiveData(arg)
            elif event == 'timer-5sec' :
                self.doGreeting(arg)
            elif event == 'shutdown' :
                self.state = 'CLOSED'
                self.doDestroyMe(arg)
            elif event == 'timer-30sec' :
                self.state = 'PING'
                self.doParentRequestRemoteID(arg)
                self.doPing(arg)
        #---CLOSED---
        elif self.state is 'CLOSED':
            pass

    def isPingOrGreeting(self, arg):
        return arg[0][:2] in [ COMMANDS['PING'], COMMANDS['GREETING'], ] 
            
    def isGreetingOrAlive(self, arg):
        return arg[0][:2] in [ COMMANDS['GREETING'], COMMANDS['ALIVE'], ]
    
    def isAlive(self, arg):
        return time.time() - self.last_alive_packet_time < 60.0 * 1.0 + 5.0
    
    def doSaveIDURL(self, arg):
        self.remote_idurl = arg
        self.remote_name = nameurl.GetName(self.remote_idurl)
  
    def doPing(self, arg):
        self.sendDatagram(COMMANDS['PING'])
    
    def doGreeting(self, arg):
        self.sendDatagram(COMMANDS['GREETING']+misc.getLocalID())
    
    def doAlive(self, arg):
        self.sendDatagram(COMMANDS['ALIVE'])
    
    def doReceiveData(self, arg):
        global _ReceiveControlFunc
        io = cStringIO.StringIO(arg[0])
        code = io.read(2)
        cmd = CODES.get(code, None)
        if cmd is None:
            # dhnio.Dprint(2, 'transport_udp_session.doReceiveData WARNING incorrect data from %s: [%s]' % (self.remote_address, code))
            return
        if _ReceiveControlFunc:
            seconds_pause = _ReceiveControlFunc(len(arg[0]))
            if seconds_pause > 0:
                # transport_udp.A().client.transport.pauseProducing()
                parent().client.client.transport.pauseProducing()
                # self.client.transport.pauseProducing()
                # reactor.callLater(seconds_pause, transport_udp.A().client.transport.resumeProducing)
                reactor.callLater(seconds_pause, parent().client.transport.resumeProducing)
        # dhnio.Dprint(8, '                                UDP:[%s] from %s' % (cmd, self.remote_address))
        self.last_alive_packet_time = time.time()
        #--- DATA
        if cmd == 'DATA':
            try:
                file_id = struct.unpack('i', io.read(4))[0]
                block_id = struct.unpack('i', io.read(4))[0]
                num_blocks = struct.unpack('i', io.read(4))[0]
                data_size = struct.unpack('i', io.read(4))[0]
            except:
                dhnio.DprintException()
                return
            if not self.incommingFiles.has_key(file_id):
                if self.receivedFiles.has_key(file_id):
                    self.sendDatagram(COMMANDS['REPORT'] + struct.pack('i', file_id) + struct.pack('i', block_id))
                    return
                # self.incommingFiles[file_id] = {}
                self.incommingFiles[file_id] = InboxFile(self, file_id, num_blocks)
            inp_data = io.read()
            if len(inp_data) != data_size:
                dhnio.Dprint(2, 'transport_udp_session.doReceiveData WARNING incorrect datagram received from %s, not full data' % str(self.remote_address))
                return
            self.incommingFiles[file_id].input_block(block_id, inp_data) 
            self.sendDatagram(COMMANDS['REPORT'] + struct.pack('i', file_id) + struct.pack('i', block_id))
            # dhnio.Dprint(10, 'transport_udp_session.doReceiveData [%d] (%d/%d) from %s' % (file_id, block_id, num_blocks, self.remote_address))
            if len(self.incommingFiles[file_id].blocks) == num_blocks:
                # dhnio.Dprint(2, 'transport_udp_session.doReceiveData new file [%d] (%d/%d) from %s' % (file_id, block_id, num_blocks, self.remote_address))
                self.incommingFiles[file_id].build()
                file_received(self.incommingFiles[file_id].filename, 'finished', 'udp', self.remote_address)
                self.incommingFiles[file_id].close()   
                del self.incommingFiles[file_id]   
                # self.incommingFiles.pop(file_id, None)
                self.receivedFiles[file_id] = time.time()
                self.eraseOldFileIDs()
        #--- REPORT
        elif cmd == 'REPORT':
            try:
                file_id = struct.unpack('i', io.read(4))[0]
            except:
                dhnio.DprintException()
                return
            if not self.outgoingFiles.has_key(file_id):
                if not self.receivedFiles.has_key(file_id):
                    dhnio.Dprint(8, 'transport_udp_session.doReceiveData WARNING unknown file_id in REPORT packet received from %s: [%d]' % (self.remote_address, file_id))
                return
            try:
                block_id = struct.unpack('i', io.read(4))[0]
            except:
                dhnio.DprintException()
                return
            self.outgoingFiles[file_id].report_block(block_id)
            self.sliding_window.pop((file_id, block_id), None)
            self.sliding_window_size += 1
            if self.sliding_window_size > MAX_WINDOW_SIZE:
                self.sliding_window_size = MAX_WINDOW_SIZE
            if len(self.outgoingFiles[file_id].blocks) == 0:
                self.reportSentDone(file_id)
                self.closeOutboxFile(file_id)
            # dhnio.Dprint(8, 'transport_udp_session.doReceiveData REPORT on [%d,%d] received, blocks=%d, window=%d' % (file_id, block_id, len(self.outgoingFiles[file_id]), self.sliding_window_size))
        #--- GREETING
        elif cmd == 'GREETING':
            greeting_idurl = io.read()
            if greeting_idurl:
                if nameurl.UrlMake(parts=nameurl.UrlParse(greeting_idurl)) != greeting_idurl:
                    dhnio.Dprint(2, 'transport_udp_session.doReceiveData WARNING incorrect idurl=%s in GREETING packet from %s' % (greeting_idurl, self.name))
                    self.sendDatagram(COMMANDS['PING'])
                else:
                    if self.remote_idurl is None:
                        self.remote_idurl = greeting_idurl
                        self.remote_name = nameurl.GetName(self.remote_idurl)
                        dhnio.Dprint(6, 'transport_udp_session.doReceiveData got idurl=%s in GREETING packet from %s' % (greeting_idurl, self.name))
                        parent().automat('recognize-remote-id', (self.index, self.remote_idurl))
                    else:
                        if self.remote_idurl != greeting_idurl:
                            dhnio.Dprint(2, 'transport_udp_session.doReceiveData WARNING wrong idurl=%s in GREETING packet from %s' % (greeting_idurl, self.name))
                    self.sendDatagram(COMMANDS['ALIVE'])
            else:
                dhnio.Dprint(2, 'transport_udp_session.doReceiveData WARNING did not found idurl in GREETING packet from %s, send PING again' % self.name)
                self.sendDatagram(COMMANDS['PING'])
        #--- PING
        elif cmd == 'PING':
            self.sendDatagram(COMMANDS['GREETING']+misc.getLocalID())
        #--- ALIVE            
        elif cmd == 'ALIVE':
            pass

    def doDestroyMe(self, arg):
        # dhnio.Dprint(6, 'transport_udp_session.doDestroyMe %s' % self.name)
        file_ids_to_remove = self.incommingFiles.keys()
        for file_id in file_ids_to_remove:
            try:
                os.close(self.incommingFiles[file_id].fd)
            except:
                dhnio.DprintException()
            file_received(self.incommingFiles[file_id].filename, 'failed', 'udp', self.remote_address, None, 'session has been closed')
            self.incommingFiles[file_id].close()   
            del self.incommingFiles[file_id]
        file_ids_to_remove = self.outgoingFiles.keys()
        for file_id in file_ids_to_remove:
            file_sent(self.remote_address, self.outgoingFiles[file_id].filename, 'failed', 'udp', None, 'session has been closed')
            self.outgoingFiles[file_id].close()
            del self.outgoingFiles[file_id]
        self.receivedFiles.clear()
        self.sliding_window.clear()
        automat.objects().pop(self.index)
        
    def doParentRequestRemoteID(self, arg):
        if self.remote_idurl:
            # dhnio.Dprint(8, 'transport_udp_session.doParentRequestRemoteID [%s] for session %s' % (self.remote_name, self.name))
            parent().automat('child-request-remote-id', self.remote_idurl)

    def doParentRecreateMe(self, arg):
        if self.remote_idurl:
            parent().automat('child-request-recreate', self.remote_idurl)


    def sendDatagram(self, data):
        # dhnio.Dprint(10, '    [UDP] %d bytes to %s (%s)' % (len(data), self.remote_address, nameurl.GetName(self.remote_idurl)))
        if parent().client is not None:
        # if transport_udp.A().client is not None:
            try:
                # transport_udp.A().client.transport.write(data, self.remote_address)
                parent().client.transport.write(data, self.remote_address)
                return True
            except:
                return False 
        else:
            dhnio.Dprint(2, 'transport_udp_session.sendDatagram WARNING client is None')
            return False      
        
    def makeFileID(self):
        return int(str(int(time.time() * 100.0))[4:])   

    def register_outbox_file(self, filename, description=''): 
        global _RegisterTransferFunc
        try:
            sz = os.path.getsize(filename)
        except:
            sz = -1
        if _RegisterTransferFunc is None:
            return None, sz
        transfer_id = _RegisterTransferFunc('send', self.remote_address, lambda:0, filename, sz, description)
        return transfer_id, sz
   
    def addOutboxFile(self, filename, description=''):
        transfer_id, filesize = self.register_outbox_file(filename, description)
        if transfer_id:
            self.outbox_queue.append((filename, filesize, transfer_id, description))
        else:
            file_sent(self.remote_address, filename, 'failed', 'udp', None, 'unable to register transfer')
        # dhnio.Dprint(6, 'transport_udp_session.addOutboxFile [%s], total outbox files: %d' % (os.path.basename(filename), len(self.outbox_queue)))
        
    def putOutboxFile(self, filename, description=''):
        transfer_id, filesize = self.register_outbox_file(filename, description)
        if transfer_id:
            self.outbox_queue.insert(0, (filename, filesize, transfer_id, description)) 
        else:
            file_sent(self.remote_address, filename, 'failed', 'udp', None, 'unable to register transfer')
        # dhnio.Dprint(6, 'transport_udp_session.putOutboxFile [%s], total outbox files: %d' % (os.path.basename(filename), len(self.outbox_queue)))
    
    def eraseOldFileIDs(self):
        if len(self.receivedFiles) > 10:
            file_ids = self.receivedFiles.keys()
            cur_tm = time.time()
            for file_id in file_ids:
                if cur_tm - self.receivedFiles[file_id] > 60 * 20:
                    del self.receivedFiles[file_id]
            del file_ids 

    def closeOutboxFile(self, file_id):
        self.outgoingFiles[file_id].close()
        del self.outgoingFiles[file_id]

    def reportSentDone(self, file_id):
        file_sent(self.remote_address, self.outgoingFiles[file_id].filename, 'finished', 'udp')
        self.block_size_level += 1
        if self.block_size_level > MAX_BLOCK_SIZE_LEVEL:
            self.block_size_level = MAX_BLOCK_SIZE_LEVEL
    
    def reportSentFailed(self, file_id, why=''):
        file_sent(self.remote_address, self.outgoingFiles[file_id].filename, 'failed', 'udp', None, why)
        self.block_size_level -= 2
        if self.block_size_level < 0:
            self.block_size_level = 0

    def processOutboxQueue(self):
        global _UnRegisterTransferFunc
        has_reads = False
        while len(self.outbox_queue) > 0 and len(self.outgoingFiles) < self.maxOutgoingFiles:        
            filename, filesize, transfer_id, description = self.outbox_queue.pop(0)
            has_reads = True
            # we have a queue of files to be sent
            # somehow file may be removed before we start sending it
            # so we check it here and skip not existed files
            if not os.path.isfile(filename):
                file_sent(self.remote_address, filename, 'failed', 'udp', None, 'file were removed')
                if _UnRegisterTransferFunc:
                    _UnRegisterTransferFunc(transfer_id)
                continue
#            if os.access(filename, os.R_OK):
#                file_sent(self.remote_address, filename, 'failed', 'udp', None, 'access denied')
#                continue
            file_id = self.makeFileID()
            f = OutboxFile(self, self.remote_idurl, file_id, filename, filesize, transfer_id, description)
            f.read_blocks()
            self.outgoingFiles[file_id] = f
        return has_reads

    def processSending(self):
        has_sends = False
        has_timedout = False
        failed_ids = []
        for file_id in self.outgoingFiles.keys():
            if self.outgoingFiles[file_id].send():
                has_sends = True
            if self.outgoingFiles[file_id].has_failed_blocks:
                failed_ids.append((file_id, 'has failed blocks'))
            elif self.outgoingFiles[file_id].is_timed_out():
                if self.outgoingFiles[file_id].timeout == 0:
                    failed_ids.append((file_id, 'canceled'))
                else:
                    failed_ids.append((file_id, 'timeout'))
                    has_timedout = True
        for file_id, why in failed_ids:
            self.reportSentFailed(file_id, why)
            self.closeOutboxFile(file_id)
        del failed_ids
        if has_timedout:
            # if some packets we currently sending is timed out
            # all other in the outbox will fail too - so erase all.
            self.clearOutboxQueue()
        self.checkWindow()
        return has_sends
    
    def clearOutboxQueue(self):
        global _UnRegisterTransferFunc
        while len(self.outbox_queue) > 0:
            filename, filesize, transfer_id, description = self.outbox_queue.pop(0)
            file_sent(self.remote_address, filename, 'failed', 'udp', None, 'timeout')
            if _UnRegisterTransferFunc:
                _UnRegisterTransferFunc(transfer_id)
            
    def checkWindow(self):
        if len(self.sliding_window) == 0:
            return
        to_remove = [] 
        to_update = {}
        cur_tm = time.time()
        for block_info, tm in self.sliding_window.items():
            if cur_tm - tm < TIME_OUT: # give some time to get an Ack. 
                continue
            to_remove.append(block_info)
            file_id = block_info[0]
            block_id = block_info[1]
            if not self.outgoingFiles.has_key(file_id):
                # dhnio.Dprint(2, 'transport_udp_session.checkWindow WARNING unknown file_id [%d] from %s' % (file_id, nameurl.GetName(self.remote_idurl)))
                continue
            if not self.outgoingFiles[file_id].blocks.has_key(block_id):
                dhnio.Dprint(2, 'transport_udp_session.checkWindow WARNING unknown block_id: %d' % block_id)
                continue
            self.outgoingFiles[file_id].send_block(block_id)
            isFailed = self.outgoingFiles[file_id].mark_failed_block(block_id)
            if not isFailed:
                to_update[block_info] = cur_tm
        for block_info in to_remove:
            del self.sliding_window[block_info]
            # dhnio.Dprint(8, 'transport_udp_session.doCheckWindow (%d,%d) removed, window=%d' % (file_id, block_id, len(self.sliding_window)))
        for block_info, tm in to_update.items():
            self.sliding_window[block_info] = tm
            # dhnio.Dprint(8, 'transport_udp_session.doCheckWindow (%s) resending to %s, window=%d' % (block_info, self.remote_address, len(self.sliding_window)))
        if len(to_remove) > 0:
            self.sliding_window_size = int( self.sliding_window_size / 2.0 )
            if self.sliding_window_size < MIN_WINDOW_SIZE:
                self.sliding_window_size = MIN_WINDOW_SIZE
            # dhnio.Dprint(8, 'transport_udp_session.doCheckWindow decreased, window=%d' % self.sliding_window_size)
        del to_remove
        del to_update

    def cancel(self, transferID):
        for i in range(len(self.outbox_queue)):
            filename, filesize, transfer_id, description = self.outbox_queue[i]
            if transferID == transfer_id:
                del self.outbox_queue[i]
                break
        for file_id in self.outgoingFiles.keys():
            if self.outgoingFiles[file_id].transfer_id == transferID:
                self.outgoingFiles[file_id].cancel()
                break
        for file_id in self.incommingFiles.keys():
            if self.incommingFiles[file_id].transfer_id == transferID:
                self.incommingFiles[file_id].build()
                file_received(self.incommingFiles[file_id].filename, 'failed', 'udp', self.remote_address)              
                self.incommingFiles[file_id].close()
                del self.incommingFiles[file_id]
                self.receivedFiles[file_id] = time.time()
                break

#------------------------------------------------------------------------------ 
    
def check_outbox_queue():    
    has_reads = False
    for sess in sessions():
        if sess.processOutboxQueue():
            has_reads = True
    return has_reads 
    
def process_outbox_queue():
    global _OutboxQueueDelay
    global _OutboxQueueTask
    
    has_reads = check_outbox_queue()

    _OutboxQueueDelay = misc.LoopAttenuation(
        _OutboxQueueDelay, 
        has_reads, 
        settings.MinimumSendingDelay(), 
        settings.MaximumSendingDelay(),)
    
    # attenuation
    _OutboxQueueTask = reactor.callLater(_OutboxQueueDelay, process_outbox_queue)
        

def process_sending():
    global _SendingDelay
    global _SendingTask
    
    has_sends = False
    for sess in sessions():
        if sess.processSending():
            has_sends = True
    
    _SendingDelay = misc.LoopAttenuation(
        _SendingDelay, 
        has_sends, 
        settings.MinimumSendingDelay(), 
        settings.MaximumSendingDelay(),)
    
    # attenuation
    _SendingTask = reactor.callLater(_SendingDelay, process_sending)
    
#------------------------------------------------------------------------------ 

def sessions():
    for a in automat.objects().values():
        if a.name.startswith('udp_'):
            yield a


def open_session(address):
    name = 'udp_%s_%s' % (address[0].replace('.','_'), address[1])
    for a in automat.objects().values():
        if a.name == name:
            return a
    a = TransportUDPSession(address, name, 'AT_STARTUP', 12)
    return a


def get_session(address):
    name = 'udp_%s_%s' % (address[0].replace('.','_'), address[1])
    for a in automat.objects().values():
        if a.name == name:
            return a
    return None


def get_session_by_index(index):
    return automat.objects().get(index, None)


def is_session_opened(address):
    name = 'udp_%s_%s' % (address[0].replace('.','_'), address[1])
    for a in automat.objects().values():
        if a.name == name:
            return True
    return False
        

def outbox_file(address, filename, fast=False, description=''):
#    dhnio.Dprint(6, 'transport_udp_session.outbox_file %s to %s (%s)' % (
#        os.path.basename(filename), nameurl.GetName(idurl), str(address)))
    s = open_session(address)
    if s.state != 'CONNECTED':
        file_sent(s.remote_address, filename, 'failed', 'udp', None, 'remote peer is not connected')
        return
    if fast:
        s.putOutboxFile(filename, description)
    else:
        s.addOutboxFile(filename, description)
    check_outbox_queue()


def cancel(transferID):
    for sess in sessions():
        sess.cancel(transferID)
    

def big_event(event, arg=None):
    for sess in sessions():
        sess.event(event, arg)


def data_received(datagram, address):
    try:
        from transport_control import black_IPs_dict
        if address[0] in black_IPs_dict().keys() or '.'.join(address[0].split('.')[:-1]) in black_IPs_dict().keys():
            # dhnio.Dprint(12, 'transport_udp_session.data_received %d bytes from BLACK IP: %s, %s' % (len(datagram), str(address[0]), str(black_IPs_dict()[address[0]])) )
            return
    except:
        pass
    if not is_session_opened(address):
        s = open_session(address)
        idurl = identitycache.GetIDURLByIPPort(address[0], address[1])
        if idurl:
            dhnio.Dprint(6, 'transport_udp_session.data_received from known contact, made a new session %s for [%s]' % (s.name, nameurl.GetName(idurl)))
            s.event('init', idurl)
        else:
            dhnio.Dprint(6, 'transport_udp_session.data_received from unknown address, made a new session %s' % s.name)
            s.event('init', None)
    A(address, 'datagram-received', (datagram, address))


def shutdown_all_sessions():
    big_event('shutdown')
    
    
def file_received(filename, status, proto='', host=None, error=None, message=''):
    global _ReceiveStatusCallback
    if _ReceiveStatusCallback is not None:
        _ReceiveStatusCallback(filename, status, proto, host, error, message)
    
    
def file_sent(host, filename, status, proto='', error=None, message=''):
    global _SendStatusCallback
    if _SendStatusCallback is not None:
        _SendStatusCallback(host, filename, status, proto, error, message)


#def timer30sec():
#    global _Timer30sec
#    big_event('timer-30sec')
#    _Timer30sec = reactor.callLater(ndom.randint(25,35), timer30sec)


def timerAlive():
    global _TimerAlive
    big_event('timer-30sec')
    _TimerAlive = reactor.callLater(30, timerAlive)


def timerPing():
    global _TimerPing
    big_event('timer-5sec')
    _TimerPing = reactor.callLater(5, timerPing)
    
#------------------------------------------------------------------------------ 
    
def SetReceiveStatusCallback(cb):
    global _ReceiveStatusCallback
    _ReceiveStatusCallback = cb
    
def SetSendStatusCallback(cb):
    global _SendStatusCallback
    _SendStatusCallback = cb
    
def SetSendControlFunc(f):
    global _SendControlFunc
    _SendControlFunc = f
    
def SetReceiveControlFunc(f):
    global _ReceiveControlFunc
    _ReceiveControlFunc = f

def SetRegisterTransferFunc(f):
    global _RegisterTransferFunc
    _RegisterTransferFunc = f

def SetUnRegisterTransferFunc(f):
    global _UnRegisterTransferFunc
    _UnRegisterTransferFunc = f

def SetStateChangedCallbackFunc(f):
    global _StateChangedCallbackFunc
    _StateChangedCallbackFunc = f
    
#------------------------------------------------------------------------------ 

    
def StartTimers():
    global _Timer30sec
    global _TimerAlive
    global _TimerPing
#    if _Timer30sec is None:
#        timer30sec()
    if _TimerAlive is None: 
        timerAlive()
    if _TimerPing is None:
        timerPing()


def StopTimers():
    global _Timer30sec
    global _TimerAlive
    global _TimerPing
#    if _Timer30sec:
#        _Timer30sec.cancel()
#        _Timer30sec = None
    if _TimerAlive:
        _TimerAlive.cancel()
        _TimerAlive = None
    if _TimerPing:
        _TimerPing.cancel()
        _TimerPing = None








from socket import socket, AF_INET, SOCK_DGRAM
from socket import error as sock_error
from nitro.tcp import tcpListen
from cspace.dht.rpc import RPCSocket
from cspace.dht.node import DHTNode
from cspace.network.router import Router

class NetworkNode( object ) :
    def __init__( self, listener, rpcSock, reactor, knownNodes=[] ) :
        self.listener = listener
        self.rpcSock = rpcSock
        self.reactor = reactor
        self.router = Router( self.listener, self.reactor )
        self.dhtNode = DHTNode( self.rpcSock, self.reactor, knownNodes )

    def getAddr( self ) :
        return self.listener.getSock().getsockname()

    def close( self ) :
        self.router.close()
        self.router = None
        self.dhtNode.close()
        self.dhtNode = None

def startNetworkNode( ipAddr, reactor, knownNodes=[], startNode=10001 ) :
    for port in xrange(startNode,startNode+9999) :
        addr = (ipAddr,port)
        try :
            listener = tcpListen( addr, reactor, None )
        except sock_error :
            continue
        sock = socket( AF_INET, SOCK_DGRAM )
        try :
            sock.bind( addr )
        except sock_error :
            sock.close()
            listener.close()
            continue
        rpcSock = RPCSocket( sock, reactor )
        node = NetworkNode( listener, rpcSock, reactor, knownNodes )
        return node

def main() :
    import sys
    from nitro.selectreactor import SelectReactor
    logging.getLogger().addHandler( logging.StreamHandler() )
    args = sys.argv[1:]
    if args :
        ipAddr = args.pop( 0 )
    else :
        ipAddr = '127.0.0.1'
    reactor = SelectReactor()
    knownNodes = []
    for nodeAddr in args :
        try :
            ip,port = nodeAddr.split( ':' )
            port = int(port)
        except TypeError, ValueError :
            logger.error( 'invalid nodeaddr: %s' % nodeAddr )
            return
        knownNodes.append( (ip,port) )
    node = startNetworkNode( ipAddr, reactor, knownNodes )
    if node is None :
        logger.error( 'unable to bind to any port on %s' % ipAddr )
        return
    logger.info( 'node address =', node.getAddr() )
    reactor.run()

if __name__ == '__main__' :
    main()



import time
from ncrypt0.x509 import makeX509Certificate, makeX509CASignedCertifiacte, cspace_ca_cert
from ncrypt0.ssl import SSLContext, SSL_METHOD_TLSv1, SSL_VERIFY_MODE_NONE, SSL_VERIFY_MODE_SELF_SIGNED

#------------------------------------------------------------------------------ 

CONTEXT_FOR_CLIENT = 1
CONTEXT_FOR_SERVER = 2

_CA_KEY_PATH = 'ca.key'
_CA_CERT_PATH = 'ca.pem'

#------------------------------------------------------------------------------ 

def set_CA_key_path(pth):
    global _CA_KEY_PATH
    _CA_KEY_PATH = pth
    
def set_CA_cert_path(pth):
    global _CA_CERT_PATH
    _CA_CERT_PATH = pth

#------------------------------------------------------------------------------ 

def makeClientSSLContext( userName, keyID, rsaKey, verify_callback=None) :
    return makeSSLContext(userName, keyID, rsaKey, verify_callback, CONTEXT_FOR_CLIENT)

def makeServerSSLContext( userName, keyID, rsaKey, verify_callback=None) :
    return makeSSLContext(userName, keyID, rsaKey, verify_callback, CONTEXT_FOR_SERVER)

#------------------------------------------------------------------------------ 

def makeSSLContext( userName, keyID, rsaKey, verify_callback=None, clientORserver=None) :
    # cert = makeX509Certificate(userName, rsaKey)
    global _CA_KEY_PATH
    global _CA_CERT_PATH
    cert = makeX509CASignedCertifiacte(userName, keyID, rsaKey, _CA_KEY_PATH, _CA_CERT_PATH)
    sslContext = SSLContext( SSL_METHOD_TLSv1 )
    if clientORserver == CONTEXT_FOR_CLIENT:
        sslContext.setPrivateKey( rsaKey )
        sslContext.setCertificate( cert )
        sslContext.checkPrivateKey()
        sslContext.setVerifyMode( SSL_VERIFY_MODE_SELF_SIGNED, verify_callback )
        sslContext.loadVerifyLocations(_CA_CERT_PATH)
    elif clientORserver == CONTEXT_FOR_SERVER:
        sslContext.setPrivateKey( rsaKey )
        sslContext.setCertificate( cert )
        sslContext.checkPrivateKey()
        sslContext.setVerifyMode( SSL_VERIFY_MODE_SELF_SIGNED, verify_callback )
        sslContext.loadVerifyLocations(_CA_CERT_PATH)
    else:
        sslContext.setPrivateKey( rsaKey )
        sslContext.setCertificate( cert )
        sslContext.checkPrivateKey()
        sslContext.setVerifyMode( SSL_VERIFY_MODE_NONE, verify_callback )
    return sslContext




import os, sys, threading, logging
from string import Template
from ncrypt0.rand import bytes as rand_bytes
from ncrypt0.rsa import RSAKey, RSAError
from nitro.selectreactor import SelectReactor
from nitro.tcp import tcpListen, TCPStream
from nitro.ssl import sslAbort
from nitro.linestream import TCPLineStream

from cspace.util.spawn import spawnProcess
from cspace.util.hexcode import hexEncode, hexDecode, HexDecodeError
from cspace.util.wordcode import wordEncode, wordDecode, WordDecodeError
from cspace.util.settings import getAppDir
from cspace.util.queue import ThreadQueue
from cspace.main.common import localSettings, appSettings, \
        isValidUserName, isValidServiceName
from cspace.main.sslbridge import SSLBridge

logger = logging.getLogger('cspace.main.appletserver')

def _substituteMetaVars( s ) :
    j = (lambda x : os.path.join(sys.prefix, x))
    prefix = sys.prefix + os.pathsep
    if sys.platform == 'win32' :
        _metaDict = dict( python=j('python.exe'), pythonw=j('pythonw.exe' ))
    else :
        _metaDict = dict( python=j('python'), pythonw=j('python') )
    _metaDict['approot'] = getAppDir()
    return Template( s ).safe_substitute( _metaDict )

def _readCommand( settings, entryPath ) :
    data = settings.getData( entryPath ).strip()
    lines = [line.strip() for line in data.split('\n')]
    return lines

class ServiceConfig( object ) :
    def _listServices( self, settings ) :
        services = []
        for entry in settings.listEntries('Services') :
            serviceName = entry.split('/')[-1]
            if not isValidServiceName(serviceName):
                continue
            serviceCommand = _readCommand( settings, entry )
            services.append( (serviceName,serviceCommand) )
        return services

    def listSystemServices( self ) :
        return self._listServices( appSettings() )

    def listUserServices( self ) :
        return self._listServices( localSettings() )

    def listActiveServices( self ) :
        sysServices = self.listSystemServices()
        userServices = self.listUserServices()
        serviceDict = {}
        out = []
        for x in userServices+sysServices :
            if x[0] in serviceDict : continue
            serviceDict[x[0]] = x
            out.append( x )
        return out

class ActionConfig( object ) :
    def _listActions( self, settings ) :
        actions = []
        for entry in settings.listEntries('ContactActions') :
            actionDir = entry.split('/')[-1]
            if not isValidServiceName(actionDir):
                continue
            actionName = settings.getData( entry+'/Action' ).strip()
            actionCommand = _readCommand( settings, entry+'/Command' )
            actionOrder = settings.getInt(entry+'/SortOrder',10000)
            actions.append( (actionDir,actionName,actionCommand,actionOrder) )
        return actions

    def listSystemActions( self ) :
        return self._listActions( appSettings() )

    def listUserActions( self ) :
        return self._listActions( localSettings() )

    def listActiveActions( self ) :
        sysActions = self.listSystemActions()
        userActions = self.listUserActions()
        actionDict = {}
        out = []
        for x in userActions+sysActions :
            if x[0] in actionDict : continue
            actionDict[x[0]] = x
            out.append( x )
        return out

class BridgeThread( threading.Thread ) :
    def __init__( self ) :
        threading.Thread.__init__( self )
        self.reactor = SelectReactor()
        self.threadQueue = ThreadQueue( self._onMessage, self.reactor )
        self.bridges = {}
        self.start()

    def _onMessage( self, msg ) :
        cmd,args = msg[0],msg[1:]
        if cmd == 'bridge' :
            sock,sslConn = args
            bridge = SSLBridge( sock, sslConn, self.reactor )
            self.bridges[bridge] = 1
            bridge.setCloseCallback( lambda : self._onBridgeClosed(bridge) )
        elif cmd == 'clear' :
            for b in self.bridges.keys() :
                b.shutdown()
            self.bridges.clear()
        elif cmd == 'stop' :
            for b in self.bridges.keys() :
                b.shutdown()
            self.bridges.clear()
            self.reactor.stop()

    def _onBridgeClosed( self, bridge ) :
        del self.bridges[bridge]

    def run( self ) :
        self.reactor.run()

class AppletConnection( object ) :
    DEFAULT = 0
    CONNECTING = 1
    WAITING_BRIDGE = 2
    LISTENER = 3
    CLOSED = 4
    def __init__( self, sock, reactor, appletServer ) :
        self.reactor = reactor
        self.stream = TCPLineStream( sock, reactor )
        self.appletServer = appletServer
        self.appletServer.appletConnections[self] = 1
        self.session = appletServer.session
        self.incoming = appletServer.incoming
        self.state = self.DEFAULT
        self._writeData = self.stream.writeData
        rt = {}
        self.requestTable = rt
        rt['echo'] = self._doEcho
        rt['getcontacts'] = self._doGetContacts
        rt['getpubkey'] = self._doGetPubKey
        rt['getcontactpubkeys'] = self._doGetContactPubKeys
        rt['connect'] = self._doConnect
        rt['connectpubkey'] = self._doConnectPubKey
        rt['accept'] = self._doAccept
        rt['getincomingpubkey'] = self._doGetIncomingPubKey
        rt['registerlistener'] = self._doRegisterListener
        rt['sendlistener'] = self._doSendListener
        self.stream.setInputCallback( self._onInput )
        self.stream.setCloseCallback( self._onClose )
        self.stream.setErrorCallback( self._onError )
        self.stream.enableRead( True )

    def _setClosed( self ) :
        del self.appletServer.appletConnections[self]
        self.state = self.CLOSED

    def shutdown( self, deferred=False ) :
        if self.state == self.CONNECTING :
            self.connectOp.cancel()
        elif self.state == self.LISTENER :
            self.appletServer.unregisterListener( self.listenerName )
        elif self.state == self.WAITING_BRIDGE :
            sslAbort( self.peerSSLConn )
        self.stream.close( deferred )
        self._setClosed()

    def _onClose( self ) :
        self.shutdown()

    def _onError( self, err, errMsg ) :
        self.shutdown()

    def _writeLine( self, line ) :
        self._writeData( line + '\r\n' )

    def _writeWords( self, words ) :
        words = [wordEncode(w) for w in words]
        self._writeData( ' '.join(words) + '\r\n' )

    def _writeError( self, msg ) :
        self._writeLine( 'ERROR %s' % msg )

    def _writeResult( self, words ) :
        self._writeWords( ['OK'] + words )

    def dispatchMessage( self, msg ) :
        assert self.state == self.LISTENER
        self._writeWords( ['MSG'] + msg )

    def _doEcho( self, words ) :
        self._writeResult( words )

    def _doGetContacts( self, words ) :
        if len(words) != 0 :
            self._writeError( 'Malformed request' )
            return
        if not self.session.isOnline() :
            self._writeError( 'Not online' )
            return
        names = self.session.getProfile().getContactNames()
        self._writeResult( names )

    def _doGetPubKey( self, words ) :
        if len(words) > 1 :
            self._writeError( 'Malformed request' )
            return
        if not self.session.isOnline() :
            self._writeError( 'Not online' )
            return
        if len(words) == 0 :
            keyData = self.session.getProfile().rsaKey.toDER_PublicKey()
            self._writeResult( [hexEncode(keyData)] )
            return
        contact = self.session.getProfile().getContactByName( words[0] )
        if contact is None :
            self._writeError( 'Unknown contact' )
            return
        self._writeResult( [hexEncode(contact.publicKeyData)] )

    def _doGetContactPubKeys( self, words ) :
        if len(words) != 0 :
            self._writeError( 'Malformed request' )
            return
        if not self.session.isOnline() :
            self._writeError( 'Not online' )
            return
        out = []
        profile = self.session.getProfile()
        for name in profile.getContactNames() :
            c = profile.getContactByName( name )
            out.extend( [c.name,hexEncode(c.publicKeyData)] )
        self._writeResult( out )

    def _connectInternal( self, publicKey, service ) :
        def onWriteComplete() :
            self.stream.shutdown()
            sock = self.stream.getSock()
            self.appletServer.bridgeThread.threadQueue.postMessage(
                    ('bridge',sock,self.peerSSLConn) )
            self._setClosed()
        def onConnect( err, sslConn ) :
            if err < 0 :
                self._writeError( 'Connect failed' )
                self.state = self.DEFAULT
                return
            self._writeResult( ['Connected'] )
            self.peerSSLConn = sslConn
            self.state = self.WAITING_BRIDGE
            self.stream.enableRead( False )
            self.stream.setWriteCompleteCallback( onWriteComplete )
        self.connectOp = self.session.connectTo( publicKey, service,
                onConnect )
        self.state = self.CONNECTING

    def _doConnect( self, words ) :
        if len(words) != 2 :
            self._writeError( 'Malformed request' )
            return
        contactName, service = words
        if not self.session.isOnline() :
            self._writeError( 'Not online' )
            return
        contact = self.session.getProfile().getContactByName( contactName )
        if not contact :
            self._writeError( 'Unknown contact' )
            return
        self._connectInternal( contact.publicKey, service )

    def _doConnectPubKey( self, words ) :
        if len(words) != 2 :
            self._writeError( 'Malformed request' )
            return
        hexPubKey, service = words
        if not self.session.isOnline() :
            self._writeError( 'Not online' )
            return
        try :
            pubKeyData = hexDecode( hexPubKey )
            pubKey = RSAKey()
            pubKey.fromDER_PublicKey( pubKeyData )
        except (HexDecodeError,RSAError) :
            self._writeError( 'Malformed publickey' )
            return
        self._connectInternal( pubKey, service )

    def _doAccept( self, words ) :
        if len(words) != 1 :
            self._writeError( 'Malformed request' )
            return
        connectionId = words[0]
        sslConn = self.incoming.acceptIncoming( connectionId )
        if not sslConn :
            self._writeError( 'Invalid connection' )
            return
        self._writeResult( ['Connected'] )
        self.peerSSLConn = sslConn
        self.state = self.WAITING_BRIDGE
        self.stream.enableRead( False )
        def onWriteComplete() :
            self.stream.shutdown()
            sock = self.stream.getSock()
            self.appletServer.bridgeThread.threadQueue.postMessage(
                    ('bridge',sock,self.peerSSLConn) )
            self._setClosed()
        self.stream.setWriteCompleteCallback( onWriteComplete )

    def _doGetIncomingPubKey( self, words ) :
        if len(words) != 1 :
            self._writeError( 'Malformed request' )
            return
        connectionId = words[0]
        peerKey = self.incoming.getPeerKey( connectionId )
        if not peerKey :
            self._writeError( 'Invalid connection' )
            return
        self._writeResult( [hexEncode(peerKey.toDER_PublicKey())] )

    def _doRegisterListener( self, words ) :
        if len(words) != 1 :
            self._writeError( 'Malformed request' )
            return
        listenerName = words[0]
        result = self.appletServer.registerListener( listenerName, self )
        if not result :
            self._writeError( 'Listener already registered' )
            return
        self.listenerName = listenerName
        self.state = self.LISTENER
        self._writeResult( ['Registered'] )

    def _doSendListener( self, words ) :
        if len(words) <= 1 :
            self._writeError( 'Malformed request' )
            return
        listenerName = words[0]
        listener = self.appletServer.getListener( listenerName )
        if listener is None :
            self._writeError( 'No such listener' )
            return
        listener.dispatchMessage( words[1:] )
        self._writeResult( ['Sent'] )

    def _onInput( self, line ) :
        assert self.state in (self.DEFAULT,self.CONNECTING,self.LISTENER)
        if self.state in (self.CONNECTING,self.LISTENER) :
            self._writeError( 'Junk received' )
            self.shutdown( deferred=True )
            return
        words = line.strip().split()
        if len(words) == 0 : return
        try :
            words = [wordDecode(w) for w in words]
        except WordDecodeError :
            self._writeError( 'Malformed request' )
            return
        cmd = words[0].lower()
        handler = self.requestTable.get( cmd, None )
        if not handler :
            self._writeError( 'Unknown request' )
            return
        handler( words[1:] )

class IncomingConnections( object ) :
    def __init__( self, reactor ) :
        self.reactor = reactor
        self.connections = {}

    def clearConnections( self ) :
        for sslConn,peerKey,timerOp in self.connections.values() :
            sslAbort( sslConn )
            timerOp.cancel()
        self.connections.clear()

    def addIncoming( self, sslConn, peerKey ) :
        while True :
            connectionId = hexEncode( rand_bytes(8) )
            if connectionId not in self.connections : break
        def onTimeout() : self._onTimeout( connectionId )
        timerOp = self.reactor.callLater( 30, onTimeout )
        self.connections[connectionId] = (sslConn,peerKey,timerOp)
        return connectionId

    def acceptIncoming( self, connectionId ) :
        info = self.connections.pop( connectionId, None )
        if info is None :
            return None
        sslConn,peerKey,timerOp = info
        timerOp.cancel()
        return sslConn

    def getPeerKey( self, connectionId ) :
        info = self.connections.get( connectionId )
        if info is None :
            return None
        return info[1]

    def _onTimeout( self, connectionId ) :
        sslConn,peerKey,timerOp = self.connections.pop( connectionId )
        sslAbort( sslConn )

class AppletServer( object ) :
    def __init__( self, session, actionManager, reactor ) :
        global logger
        self.session = session
        self.actionManager = actionManager
        self.reactor = reactor
        self.listener = tcpListen( ('127.0.0.1',0), reactor, self._onNewConnection )
        self.listenPort = self.listener.getSock().getsockname()[1]
        logger.info('appletserver listenport = %d' % self.listenPort)
        self.serviceConfig = ServiceConfig()
        self.actionConfig = ActionConfig()
        self.listeners = {}
        self.incoming = IncomingConnections( self.reactor )
        self.services = []
        self.actions = []
        self.appletConnections = {}
        self.onServiceCallback = None
        self.onActionCallback = None
        self.bridgeThread = BridgeThread()
        
    def registerServices(self):
        for (service,command) in self.serviceConfig.listActiveServices() :
            def doRegisterService( service, command ) :
                def onService( sslConn, peerKey, contactName, incomingName ) :
                    self._onService( service, command, sslConn,
                            peerKey, contactName, incomingName )
                self.session.registerService( service, onService )
            doRegisterService( service, command )
            self.services.append( service )
    
    def registerActions(self):
        for (actionDir,action,command,order) in self.actionConfig.listActiveActions() :
            def doRegisterAction( actionDir, action, command, order ) :
                def onAction( contactName ) :
                    self._onAction( actionDir, action, command, contactName )
                return self.actionManager.registerAction( action, onAction, order )
            actionId = doRegisterAction( actionDir, action, command, order )
            self.actions.append( actionId )
            if actionDir == 'TextChat' :
                self.actionManager.setDefaultAction( actionId )

    def shutdown( self ) :
        self.incoming.clearConnections()
        appletConns = self.appletConnections.keys()
        for conn in appletConns :
            conn.shutdown()
        self.bridgeThread.threadQueue.postMessage( ('stop',) )
        self.bridgeThread.join()
        self.listener.close()

    def clearConnections( self ) :
        self.incoming.clearConnections()
        appletConns = self.appletConnections.keys()
        for conn in appletConns :
            conn.shutdown()
        self.bridgeThread.threadQueue.postMessage( ('clear',) )

    def getListenPort( self ) : return self.listenPort

    def registerListener( self, name, connection ) :
        conn = self.listeners.setdefault( name, connection )
        return conn is connection

    def unregisterListener( self, name ) :
        del self.listeners[name]

    def getListener( self, name ) :
        return self.listeners.get( name, None )

    def _onNewConnection( self, sock ) :
        AppletConnection( sock, self.reactor, self )

    def _findProgram( self, relPath ) :
        dirList = os.environ.get( 'PATH', '' ).split( ';' )
        for d in dirList :
            p = os.path.join( d, relPath )
            if os.path.isfile(p) :
                return p
        return relPath

    def _runCommand( self, command, envNew ) :
        env = dict( os.environ.items() )
        env.update( envNew )
        cmdLine = [_substituteMetaVars(x) for x in command]
        p = os.path.join( getAppDir(), cmdLine[0] )
        if not os.path.isfile(p) :
            p = self._findProgram( cmdLine[0] )
        args = [p] + cmdLine[1:]
        startingDir = os.getcwd()
        result = spawnProcess( p, args, env, startingDir, 0 )
        if not result :
            global logger
            logger.error( 'error starting command (%s)' % p )

    def _onService( self, service, command, sslConn, peerKey,
            contactName, incomingName ) :
        connectionId = self.incoming.addIncoming( sslConn, peerKey )
        if self.onServiceCallback:
            self.onServiceCallback(service, command, sslConn, peerKey, contactName, incomingName, self.listenPort)
            return
        env = {}
        env['CSPACE_PORT'] = str(self.listenPort)
        env['CSPACE_USER'] = self.session.getProfile().name
        env['CSPACE_EVENT'] = 'INCOMING'
        env['CSPACE_SERVICE'] = service
        env['CSPACE_CONNECTIONID'] = connectionId
        env['CSPACE_CONTACTNAME'] = contactName
        env['CSPACE_INCOMINGNAME'] = incomingName
        self._runCommand( command, env )

    def _onAction( self, actionDir, action, command, contactName ) :
        if self.onActionCallback:
            self.onActionCallback(actionDir, action, command, contactName, self.listenPort)
            return
        env = {}
        env['CSPACE_PORT'] = str(self.listenPort)
        env['CSPACE_USER'] = self.session.getProfile().name
        env['CSPACE_EVENT'] = 'CONTACTACTION'
        env['CSPACE_CONTACTNAME'] = contactName
        env['CSPACE_ACTIONDIR'] = actionDir
        env['CSPACE_ACTION'] = action
        logger.info( str(env) )
        self._runCommand( command, env )



import logging
from time import time
from random import choice, randint
from nitro.async import AsyncOp
from cspace.dht.params import DHT_K, DHT_ID_MAX
from cspace.dht.util import idToNum, numToId, addrToId
from cspace.dht.util import verifySignature
from cspace.dht.proto import MESSAGES, validateRequest, ProtocolError
from cspace.dht.client import DHTClient
from cspace.dht.firewalltest import firewallTestClient

DEFAULT_DATA_TIMEOUT = 15*60
STALE_DATA_INTERVAL = 10
REFRESH_NODES_INTERVAL = 10*60

logger = logging.getLogger( 'cspace.dht.node' )

class Dummy : pass

class DataStore( object ) :
    def __init__( self ) :
        self.d = {}

    def set( self, publicKeyData, data, updateLevel,
            signature, timeout ) :
        deadline = time() + timeout
        self.d[publicKeyData] = (data,updateLevel,signature,deadline)

    def get( self, publicKeyData ) :
        entry = self.d.get( publicKeyData )
        if entry is None :
            return None
        return entry[:3]

    def removeStale( self ) :
        t = time()
        for k,x in self.d.items() :
            if x[-1] < t : del self.d[k]

class KNodeInfo( object ) :
    def __init__( self, nodeAddr, nodeNumId ) :
        self.nodeAddr = nodeAddr
        self.nodeNumId = nodeNumId
        self.lastSeen = time()

    def refresh( self ) :
        self.lastSeen = time()

class KBucket( object ) :
    def __init__( self, start, end ) :
        self.start = start
        self.end = end
        self.nodes = []
        self.nodeDict = {}

    def size( self ) : return len(self.nodes)

    def contains( self, numId ) :
        return (self.start <= numId < self.end)

    def removeNode( self, nodeNumId ) :
        del self.nodeDict[nodeNumId]
        for i,nodeInfo in enumerate(self.nodes) :
            if nodeInfo.nodeNumId == nodeNumId : break
        del self.nodes[i]

    def addNode( self, nodeAddr, nodeNumId ) :
        assert self.contains( nodeNumId )
        if nodeNumId not in self.nodeDict :
            nodeInfo = KNodeInfo( nodeAddr, nodeNumId )
            self.nodes.insert( 0, nodeInfo )
            self.nodeDict[nodeNumId] = 1
            displaced = self.nodes[DHT_K:]
            del self.nodes[DHT_K:]
            for nodeInfo in displaced :
                del self.nodeDict[nodeInfo.nodeNumId]
        else :
            for i,nodeInfo in enumerate(self.nodes) :
                if nodeInfo.nodeNumId == nodeNumId : break
            del self.nodes[i]
            nodeInfo.refresh()
            self.nodes.insert( 0, nodeInfo )

    def split( self ) :
        mid = (self.start + self.end)/2
        a = KBucket( self.start, mid )
        b = KBucket( mid, self.end )
        for nodeInfo in self.nodes :
            c = a.contains(nodeInfo.nodeNumId) and a or b
            c.nodes.append( nodeInfo )
            c.nodeDict[nodeInfo.nodeNumId] = 1
        return (a,b)

class KTable( object ) :
    def __init__( self, selfNumId ) :
        self.selfNumId = selfNumId
        self.table = [KBucket(0,DHT_ID_MAX)]

    def nodeSeen( self, nodeAddr, nodeNumId=None ) :
        if nodeNumId is None :
            nodeNumId = idToNum( addrToId(nodeAddr) )
        if nodeNumId == self.selfNumId : return
        for i,bucket in enumerate(self.table) :
            if bucket.contains(nodeNumId) : break
        if not bucket.contains(self.selfNumId) :
            bucket.addNode( nodeAddr, nodeNumId )
        else :
            self.table[i:i+1] = bucket.split()
            self.nodeSeen( nodeAddr, nodeNumId )

    def getClosestNodes( self, destNumId ) :
        for i,bucket in enumerate(self.table) :
            if bucket.contains(destNumId) :
                break
        output = bucket.nodes[:]
        remaining = DHT_K - len(output)
        if remaining :
            def fetchLeft( i, n ) :
                if (n == 0) or (i == 0) : return []
                out = self.table[i-1].nodes[:]
                return out + fetchLeft(i-1,n-len(out))
            def fetchRight( i, n ) :
                if (n == 0) or (i+1 == len(self.table)) : return []
                out = self.table[i+1].nodes[:]
                return out + fetchRight(i+1,n-len(out))
            output.extend( fetchLeft(i,remaining) )
            output.extend( fetchRight(i,remaining) )
            output.sort( key = lambda ni : ni.nodeNumId^destNumId )
            del output[DHT_K:]
            output.sort( key = lambda ni : ni.lastSeen, reverse=True )
        return [ni.nodeAddr for ni in output]

class KTracker( object ) :
    def __init__( self, ktable ) :
        self.ktable = ktable

    def nodeSeen( self, nodeAddr, latency ) :
        self.ktable.nodeSeen( nodeAddr )

    def nodeFailed( self, nodeAddr ) :
        pass

class DHTNode( object ) :
    def __init__( self, rpcSocket, reactor, knownNodes=[] ) :
        self.rpcSocket = rpcSocket
        self.reactor = reactor
        self.rpcSocket.setRequestCallback( self._onInput )
        self.store = DataStore()
        self.ktable = KTable( idToNum(addrToId(rpcSocket.getAddr())) )
        self.ktracker = KTracker( self.ktable )
        for nodeAddr in knownNodes :
            self.ktable.nodeSeen( nodeAddr )
        self.client = DHTClient( rpcSocket, False, self.ktracker )
        self.requestTable = {}
        for x in MESSAGES :
            self.requestTable[x] = getattr( self, 'do%s' % x )
        self.otherOps = set()
        self._initNodeRefresher()
        self._initDataTimer()

    def close( self ) :
        for op in self.otherOps :
            op.cancel()
        self.otherOps.clear()
        self.rpcSocket.close()
        self.rpcSocket = None

    def _onInput( self, msg, ctx ) :
        try :
            cmd,useSourceAddr,params = validateRequest( msg )
        except ProtocolError, e :
            logger.exception( 'invalid request msg' )
            return
        if useSourceAddr :
            self.ktable.nodeSeen( ctx.addr )
        self.requestTable[cmd]( msg, ctx )

    def doPing( self, msg, ctx ) :
        ctx.response( '' )

    def doGetAddr( self, msg, ctx ) :
        ctx.response( ctx.addr )

    def doGetKey( self, msg, ctx ) :
        publicKeyData = msg[0]
        entry = self.store.get( publicKeyData )
        if entry is None :
            ctx.response( (-1,'',0,'') )
            return
        data,updateLevel,signature = entry
        ctx.response( (0,data,updateLevel,signature) )

    def doPutKey( self, msg, ctx ) :
        publicKeyData,data,updateLevel,signature = msg
        if not verifySignature(publicKeyData,data,updateLevel,signature) :
            return
        entry = self.store.get( publicKeyData )
        if entry is not None :
            oldData,oldUpdateLevel,oldSignature = entry
            if updateLevel <= oldUpdateLevel :
                ctx.response( (-1,oldData,oldUpdateLevel,oldSignature) )
                return
        self.store.set( publicKeyData, data, updateLevel, signature,
                DEFAULT_DATA_TIMEOUT )
        ctx.response( (0,data,updateLevel,signature) )

    def doFindNodes( self, msg, ctx ) :
        destNumId = idToNum( msg[0] )
        ctx.response( self.ktable.getClosestNodes(destNumId) )

    def doFirewallCheck( self, msg, ctx ) :
        addr = tuple( msg )
        if ctx.addr != addr :
            ctx.response( (-1,'') )
            return
        def onTest( err, token ) :
            self.otherOps.remove( op )
            ctx.response( (err,token) )
        op = firewallTestClient( addr, self.reactor, onTest )
        self.otherOps.add( op )

    def _initNodeRefresher( self ):
        obj = Dummy()
        def doWait() :
            timeout = REFRESH_NODES_INTERVAL
            obj.op = self.reactor.callLater( timeout, doRefresh )
        def onLookup( nodes ) :
            doWait()
        def doRefresh() :
            bucket = choice( self.ktable.table )
            destNumId = randint( bucket.start, bucket.end-1 )
            destId = numToId( destNumId )
            startNodes = self.ktable.getClosestNodes( destNumId )
            if not startNodes :
                doWait()
                return
            obj.op = self.client.lookup( destId, startNodes, onLookup )
        def doCancel() :
            obj.op.cancel()
        obj.op = self.reactor.callLater( 0, doRefresh )
        op = AsyncOp( None, doCancel )
        self.otherOps.add( op )

    def _initDataTimer( self ) :
        def onTimer() :
            self.store.removeStale()
        timeout = STALE_DATA_INTERVAL
        timerOp = self.reactor.addTimer( timeout, onTimer )
        self.otherOps.add( timerOp )



from cspace.dht.params import DHT_ID_LENGTH, DHT_K
from cspace.dht.util import checkAddr

class ProtocolError(Exception) : pass

def _assert( cond, msg=None ) :
    if not cond :
        if msg is None : msg = 'invalid message data'
        raise ProtocolError, msg

def validatePingRequest( msg ) :
    _assert( len(msg) == 0 )
    return msg

def validatePingResponse( msg ) :
    _assert( type(msg) is str )
    _assert( len(msg) == 0 )
    return msg

def validateGetAddrRequest( msg ) :
    _assert( len(msg) == 0 )
    return msg

def validateGetAddrResponse( msg ) :
    _assert( type(msg) is list )
    msg = tuple(msg)
    _assert( checkAddr(msg) )
    return msg

def validateGetKeyRequest( msg ) :
    _assert( len(msg) == 1 )
    _assert( type(msg[0]) is str )
    return msg

def validateGetKeyResponse( msg ) :
    _assert( type(msg) is list )
    _assert( len(msg) == 4 )
    result,data,updateLevel,signature = msg
    _assert( type(result) is int )
    _assert( type(data) is str )
    _assert( type(updateLevel) is int )
    _assert( updateLevel >= 0 )
    _assert( type(signature) is str )
    return msg

def validatePutKeyRequest( msg ) :
    _assert( len(msg) == 4 )
    publicKeyData,data,updateLevel,signature = msg
    _assert( type(publicKeyData) is str )
    _assert( type(data) is str )
    _assert( type(updateLevel) is int )
    _assert( updateLevel > 0 )
    _assert( type(signature) is str )
    return msg

def validatePutKeyResponse( msg ) :
    _assert( type(msg) is list )
    _assert( len(msg) == 4 )
    result,data,updateLevel,signature = msg
    _assert( type(result) is int )
    _assert( type(data) is str )
    _assert( type(updateLevel) is int )
    _assert( updateLevel >= 0 )
    _assert( type(signature) is str )
    return msg

def validateFindNodesRequest( msg ) :
    _assert( len(msg) == 1 )
    _assert( type(msg[0]) is str )
    _assert( len(msg[0]) == DHT_ID_LENGTH )
    return msg

def validateFindNodesResponse( msg ) :
    _assert( type(msg) is list )
    msg = msg[:DHT_K]
    out = []
    for x in msg :
        _assert( type(x) is list )
        x = tuple(x)
        _assert( checkAddr(x) )
        out.append( x )
    return out

def validateFirewallCheckRequest( msg ) :
    _assert( len(msg) == 2 )
    addr = tuple( msg )
    _assert( checkAddr(addr) )
    return addr

def validateFirewallCheckResponse( msg ) :
    _assert( type(msg) is list )
    fwResult,token = msg
    _assert( type(fwResult) is int )
    _assert( type(token) is str )
    return msg

MESSAGES = ( 'Ping', 'GetAddr', 'GetKey', 'PutKey', 'FindNodes', 'FirewallCheck' )
requestValidators = {}
responseValidators = {}
def _initTables() :
    g = globals()
    for m in MESSAGES :
        requestValidators[m] = g[ 'validate' + m + 'Request' ]
        responseValidators[m] = g[ 'validate' + m + 'Response' ]
_initTables()

def validateRequest( msg ) :
    _assert( type(msg) is list )
    _assert( len(msg) >= 2 )
    cmd, useSourceAddr = msg[0], msg[1]
    del msg[:2]
    _assert( type(cmd) is str )
    _assert( type(useSourceAddr) is int )
    _assert( useSourceAddr in (0,1) )
    validator = requestValidators.get( cmd )
    _assert( validator is not None )
    msg = validator( msg )
    return (cmd, useSourceAddr, msg)

def validateResponse( cmd, msg ) :
    validator = responseValidators[cmd]
    return validator( msg )



import StringIO
import logging
from socket import socket, AF_INET, SOCK_DGRAM
from nitro.async import AsyncOp
from nitro.ssl import sslAbort
from cspace.util.validate import validateInetAddress
from cspace.util.statemachine import StateMachine
from cspace.dht.rpc import RPCSocket
from cspace.dht.client import DHTClient
from cspace.network.nodetable import NodeTable, NodeTableRefresher
from cspace.network.nodetable import initNodeTable
from cspace.network.location import lookupUser
from cspace.network.locationcache import LocationCache
from cspace.network.listener import BaseIntStore
from cspace.network.listener import UserListener, acceptIncoming
from cspace.network.connector import userServiceConnect
from cspace.network.noderunner import NodeRunner
from cspace.main.common import localSettings, profileSettings, \
        isValidUserName, isValidServiceName
from cspace.main.permissions import Permissions
from cspace.main.permissions import ACCESS_ALLOW, ACCESS_DENY, ACCESS_PROMPT

logger = logging.getLogger('cspace.main.session')

USERLIST_PUBLISH_TIMEOUT = 5*60

class Dummy : pass

class UpdateLevelStore( BaseIntStore ) :
    def __init__( self, profile ) :
        self.entry = profile.storeEntry + '/UpdateLevel'

    def get( self ) :
        return profileSettings().getInt( self.entry, 1 )

    def set( self, value ) :
        return profileSettings().setInt( self.entry, value )

def _loadNodeCache( nodeTable ) :
    data = localSettings().getData( 'NodeCache', '' )
    for addr in data.split() :
        try :
            ip,port = addr.split( ':' )
            port = int(port)
            nodeAddr = (ip,port)
            validateInetAddress( nodeAddr )
        except :
            continue
        nodeTable.addNode( nodeAddr )

def _saveNodeCache( nodeTable ) :
    out = StringIO.StringIO()
    for addr in nodeTable.getAllNonSeedNodes() :
        print>>out, '%s:%d' % addr
    return localSettings().setData( 'NodeCache', out.getvalue() )

class UserSession( object ) :
    (OFFLINE,CONNECTING,ONLINE,DISCONNECTING) = range(4)

    def __init__( self, seedNodes, reactor ) :
        self.seedNodes = seedNodes
        self.reactor = reactor
        self.sm = StateMachine( self.OFFLINE )
        self.services = {}
        self.nodeTable = None
        self.nodeTableRefresher = None
        self.rpcSocket = None
        self.dhtClient = None
        self.locationCache = None
        self.profile = None
        self.listener = None
        self.permissions = None
        #self.nodeRunner = None

    def getProfile( self ) : return self.profile
    def getPermissions( self ) : return self.permissions

    def isOnline( self ) : return self.sm.current() == self.ONLINE

    def registerService( self, serviceName, serviceCallback ) :
        assert isValidServiceName(serviceName)
        if serviceName in self.services :
            return False
        self.services[serviceName] = serviceCallback
        return True

    def unregisterService( self, serviceName ) :
        if self.services.pop(serviceName,None) is None :
            return False
        return True

    def goOnline( self, profile ) :
        assert self.sm.current() == self.OFFLINE
        obj = Dummy()
        def doCleanup() :
            if obj.initNodeTableOp is not None :
                obj.initNodeTableOp.cancel()
            if obj.nodeTableRefresher is not None :
                obj.nodeTableRefresher.close()
            if obj.listener is not None :
                obj.listener.close()
            obj.rpcSocket.close()
        def onOffline() :
            #self.nodeRunner.close()
            #self.nodeRunner = None
            self.listener.close()
            self.listener = None
            self.permissions = None
            self.locationCache.close()
            self.locationCache = None
            self.nodeTableRefresher.close()
            self.nodeTableRefresher = None
            _saveNodeCache( self.nodeTable )
            self.nodeTable = None
            self.dhtClient = None
            self.rpcSocket.close()
            self.rpcSocket = None
            self.profile = None
        def onListenerStart( result ) :
            if not result :
                self.sm.removeCallback( obj.callbackId )
                doCleanup()
                self.sm.change( self.OFFLINE )
                return
            obj.listener.setCloseCallback( self._onListenerClose )
            obj.listener.setIncomingCallback( self._onIncoming )
            self.profile = profile
            self.rpcSocket = obj.rpcSocket
            self.dhtClient = obj.dhtClient
            self.nodeTable = obj.nodeTable
            self.nodeTableRefresher = obj.nodeTableRefresher
            self.locationCache = LocationCache( self.dhtClient,
                    self.nodeTable, self.reactor )
            self.listener = obj.listener
            self.permissions = Permissions( profile, self.services.keys() )
            if self.permissions.isModified() :
                self.permissions.savePermissions()
            #self.nodeRunner = NodeRunner( self.nodeTable, self.reactor )
            self.sm.removeCallback( obj.callbackId )
            self.sm.appendCallback( onOffline, dest=self.OFFLINE, single=True )
            self.sm.change( self.ONLINE )
        def onNodeTableInit() :
            obj.initNodeTableOp = None
            obj.nodeTableRefresher = NodeTableRefresher(
                    obj.nodeTable, obj.dhtClient, self.reactor )
            updateLevelStore = UpdateLevelStore( profile )
            obj.listener = UserListener( profile.name, profile.keyId, 
                    profile.rsaKey, updateLevelStore,
                    obj.dhtClient, obj.nodeTable, self.reactor )
            obj.listener.start( onListenerStart )
        self.sm.change( self.CONNECTING )
        obj.callbackId = self.sm.insertCallback( doCleanup,
                src=self.CONNECTING, single=True )
        obj.nodeTable = NodeTable( self.seedNodes )
        #_loadNodeCache( obj.nodeTable )
        udpSock = socket( AF_INET, SOCK_DGRAM )
        udpSock.bind( ('',0) )
        obj.rpcSocket = RPCSocket( udpSock, self.reactor )
        obj.dhtClient = DHTClient( obj.rpcSocket,
                nodeTracker=obj.nodeTable )
        obj.initNodeTableOp = initNodeTable( obj.nodeTable,
                obj.dhtClient, self.reactor, onNodeTableInit )
        obj.nodeTableRefresher = None
        obj.listener = None

    def _onListenerClose( self ) :
        self.sm.change( self.OFFLINE )

    def _rejectIncoming( self, sslConn, callback=None ) :
        def onReject( sslConn ) :
            if sslConn is None :
                op.notify( False )
                return
            sslAbort( sslConn )
            op.notify( True )
        rejectOp = acceptIncoming( False, sslConn, self.reactor,
                onReject )
        op = AsyncOp( callback, rejectOp.cancel )
        return op

    def _acceptIncoming( self, sslConn, serviceName, peerKey,
            contactName, peerKeyID, incomingName, callback=None ) :
        def onAccept( sslConn ) :
            if sslConn is None :
                op.notify( False )
                return
            serviceCallback = self.services.get( serviceName )
            if serviceCallback is None :
                sslAbort( sslConn )
                op.notify( False )
                return
            serviceCallback( sslConn, peerKey, contactName, peerKeyID,
                    incomingName )
            op.notify( True )
        acceptOp = acceptIncoming( True, sslConn, self.reactor,
                onAccept )
        op = AsyncOp( callback, acceptOp.cancel )
        return op

    # def _promptIncoming( self, sslConn, serviceName, peerKey,
    #         contactName, incomingName, callback=None ) :
    #     def doCancel() :
    #         promptOp.cancel()
    #         sslAbort( sslConn )
    #     def onPromptResult( promptResult ) :
    #         if not promptResult :
    #             rejectOp = self._rejectIncoming( sslConn, op.notify )
    #             op.setCanceler( rejectOp.cancel )
    #         else :
    #             acceptOp = self._acceptIncoming( sslConn, serviceName,
    #                     peerKey, contactName, incomingName, op.notify )
    #             op.setCanceler( acceptOp.cancel )
    #     promptName = contactName
    #     if not promptName :
    #         promptName = '(Unknown %s)' % incomingName
    #     promptOp = IncomingPromptWindow( promptName, serviceName,
    #             self.reactor, onPromptResult ).getOp()
    #     op = AsyncOp( callback, doCancel )
    #     return op

    def _onIncoming( self, sslConn, peerKey, incomingName, peerKeyID,
            serviceName ) :
        assert self.sm.current() == self.ONLINE
        if not isValidUserName(incomingName) :
            sslAbort( sslConn )
            return
        if not isValidServiceName(serviceName) :
            sslAbort( sslConn )
            return
        contact = self.profile.getContactByPublicKey( peerKey )
        if contact is None :
            contactName = peerKeyID
        else :
            contactName = contact.name
        if serviceName not in self.services :
            action = ACCESS_DENY
        else :
            action = self.permissions.execute( contactName,
                    serviceName )
        def onActionDone( result ) :
            self.sm.removeCallback( callbackId )
        if action == ACCESS_DENY :
            actionOp = self._rejectIncoming( sslConn, onActionDone )
        else:
        # elif action == ACCESS_ALLOW  :
            actionOp = self._acceptIncoming( sslConn, serviceName,
                    peerKey, contactName, peerKeyID, incomingName, onActionDone )
        # else :
        #     assert action == ACCESS_PROMPT
        #     actionOp = self._promptIncoming( sslConn, serviceName,
        #             peerKey, contactName, incomingName, onActionDone )
        callbackId = self.sm.insertCallback( actionOp.cancel,
                src=self.ONLINE, single=True )

    def goOffline( self ) :
        assert self.sm.current() == self.ONLINE
        def onStop( result ) :
            self.sm.change( self.OFFLINE )
        self.sm.change( self.DISCONNECTING )
        self.listener.stop( onStop )

    def shutdown( self ) :
        self.sm.change( self.OFFLINE )

    def probeUserOnline( self, publicKey, callback=None ) :
        assert self.sm.current() == self.ONLINE
        def onStateChange() :
            lookupOp.cancel()
            op.notify( False )
        def doCancel() :
            lookupOp.cancel()
            self.sm.removeCallback( callbackId )
        def onLookup( location ) :
            self.sm.removeCallback( callbackId )
            if location is None :
                op.notify( False )
                return
            if (not location.directLocations) and (not location.routedLocations) :
                op.notify( False )
                return
            op.notify( True )
        lookupOp = self.locationCache.refreshUser( publicKey,
                onLookup )
        callbackId = self.sm.insertCallback( onStateChange,
                src=self.ONLINE, single=True )
        op = AsyncOp( callback, doCancel )
        return op

    def connectTo( self, publicKey, serviceName, callback=None ) :
        assert self.sm.current() == self.ONLINE
        def onStateChange() :
            connectOp.cancel()
            op.notify( -1, None )
        def doCancel() :
            connectOp.cancel()
            self.sm.removeCallback( callbackId )
        def onConnect( sslConn ) :
            self.sm.removeCallback( callbackId )
            if sslConn is None :
                op.notify( -1, None )
                return
            op.notify( 0, sslConn )
        connectOp = userServiceConnect( self.profile.name, self.profile.keyId,
                self.profile.rsaKey, self.listener.getPublicIP(),
                publicKey, serviceName, self.locationCache,
                self.reactor, onConnect )
        callbackId = self.sm.insertCallback( onStateChange,
                src=self.ONLINE, single=True )
        op = AsyncOp( callback, doCancel )
        return op



#!/usr/bin/python

#
#    Copyright DataHaven.NET LTD. of Anguilla, 2006-2013
#    Use of this software constitutes acceptance of the Terms of Use
#      http://datahaven.net/terms_of_use.html
#    All rights reserved.
#
#  For most of DHN code (outside ssh I think) we will only use ascii encoded string versions of keys.
#  Expect to make keys, signatures, and hashes all base64 strings soon.
#
#  Our local key is always onhand.
#
#  Todo:
#     Main thing is to be able to use public keys in contacts to verify dhnpackets
#     We never want to bother storing bad data, and need localtester to do localscrub
#
# Crypto info:
# http://www.amk.ca/python/writing/pycrypt/pycrypt.html

# run several times from interpreter and it does not like loading RSA key from previous run PREPRO


import os
import random
#import base64
import hashlib
#import time

from Crypto.PublicKey import RSA
from Crypto.Cipher import DES3
#from Crypto.Cipher import AES

import warnings
warnings.filterwarnings('ignore',category=DeprecationWarning)
from twisted.conch.ssh import keys
#from twisted.spread import banana,jelly

import settings
import dhnio
import misc



# Global for this file
MyRsaKey = None 
# This will be an object
LocalKey = None

#------------------------------------------------------------------------------ 

def InitMyKey(keyfilename=None):
    global MyRsaKey
    global LocalKey
    if LocalKey is not None:
        return
    if MyRsaKey is not None:
        return
    if keyfilename is None:
        keyfilename = settings.KeyFileName()
    if os.path.exists(keyfilename+'_location'):
        newkeyfilename = dhnio.ReadTextFile(keyfilename+'_location').strip()
        if os.path.exists(newkeyfilename):
            keyfilename = newkeyfilename
    if os.path.exists(keyfilename):
        dhnio.Dprint(4, 'dhncrypto.InitMyKey load private key from %s' % keyfilename)
        LocalKey = keys.Key.fromFile(keyfilename)
        MyRsaKey = LocalKey.keyObject
    else:
        dhnio.Dprint(4, 'dhncrypto.InitMyKey generate new private key')
        # want 2048 but might have had problem - had shorter when using banana for serialization
        MyRsaKey = RSA.generate(settings.getPrivateKeySize(), os.urandom)       
        LocalKey = keys.Key(MyRsaKey)
        keystring = LocalKey.toString('openssh')
        dhnio.WriteFile(keyfilename, keystring)

def ForgetMyKey():
    global LocalKey
    global MyRsaKey
    LocalKey = None
    MyRsaKey = None

def isMyLocalKeyReady():
    global LocalKey
    return LocalKey is not None

def MyPublicKey():
    global LocalKey
    InitMyKey()
    Result = LocalKey.public().toString('openssh')
    return Result

def MyPrivateKey():
    global LocalKey
    InitMyKey()
    return LocalKey.toString('openssh')

def MyPublicKeyObject():
    global LocalKey
    InitMyKey()
    return LocalKey.public()

def MyPrivateKeyObject():
    global LocalKey
    InitMyKey()
    return LocalKey

def Sign(inp):
    #dhnio.Dprint(18, 'dhncrypto.Sign %d bytes' % len(inp))
    global LocalKey
    InitMyKey()
    # Makes a list but we just want a string
    Signature = LocalKey.keyObject.sign(inp, '')
    # so we take first element in list - need str cause was long    
    result = str(Signature[0]) 
    return result

# key is public key in string format - as is dhncrypto standard, mostly
def VerifySignature(keystring, hashcode, signature):
    keyobj = keys.Key.fromString(keystring).keyObject
    # needs to be a long in a list
    sig2 = long(signature),
    Result = bool(keyobj.verify(hashcode, sig2))
    return Result

def Verify(ConIdentity, hashcode, signature):
    key = ConIdentity.publickey
    Result = VerifySignature(key, hashcode, signature)
    return Result

def HashSHA(inp):
    return hashlib.sha1(inp).digest()
##    return(sha.new(input).digest())

def HashMD5(inp):
    return hashlib.md5(inp).digest()
##    return(md5.new(input).digest())

def Hash(inp):
    return HashMD5(inp)

#  Outside of this file we just use the string version of the public keys
def EncryptStringPK(publickeystring, inp):
    keyobj = keys.Key.fromString(publickeystring)
    return EncryptBinaryPK(keyobj, inp)

# This is just using local key
def EncryptLocalPK(inp):
    global LocalKey
    InitMyKey()
    return EncryptBinaryPK(LocalKey, inp)

# There is a bug in rsa.encrypt if there is a leading '\0' in the string.
# Only think we encrypt is produced by NewSessionKey() which takes care not to have leading zero.
# See   bug report in http://permalink.gmane.org/gmane.comp.python.cryptography.cvs/217
# So we add a 1 in front.
def EncryptBinaryPK(publickey, inp):
    atuple = publickey.keyObject.encrypt("1"+inp,"")
    #why "1"+input ???????? TODO
    #we just use strings here 
    return atuple[0]                     

# we only decrypt with our local private key so no argument for that
def DecryptLocalPK(inp):
    global MyRsaKey
    global LocalKey
    InitMyKey()
    atuple = (inp,)
    padresult = MyRsaKey.decrypt(atuple)
    result = padresult[1:]                   # remove the "1" added in EncryptBinaryPK
    return result

def SessionKeyType():
    return "DES3"

def NewSessionKey():
    return chr(random.randint(1, 255)) + os.urandom(23)   # to work around bug in rsa.encrypt
##    return(os.urandom(24))          # really random string for making equivalent DES3 objects when needed

def EncryptWithSessionKey(rand24, inp):
    SessionKey = DES3.new(rand24)
    data = misc.RoundupString(inp, 24)
    ret = SessionKey.encrypt(data)
    del data
    return ret

def DecryptWithSessionKey(rand24, inp):
    SessionKey = DES3.new(rand24)
    return SessionKey.decrypt(inp)

